from typing import Any, Dict


class SingletonMeta(type):
    _instances: Dict[Any, Any] = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super().__call__(*args, **kwargs)
        return cls._instances[cls]


class Singleton(metaclass=SingletonMeta):
    pass



from typing import Any, List, Optional

import openai

from vocode import getenv
from vocode.turn_based.agent.base_agent import BaseAgent


class ChatGPTAgent(BaseAgent):
    def __init__(
        self,
        system_prompt: str,
        api_key: Optional[str] = None,
        initial_message: Optional[str] = None,
        model_name: str = "gpt-3.5-turbo",
        temperature: float = 0.7,
        max_tokens: int = 100,
    ):
        super().__init__(initial_message=initial_message)
        openai.api_key = getenv("OPENAI_API_KEY", api_key)
        if not openai.api_key:
            raise ValueError("OpenAI API key not provided")
        self.client = openai.OpenAI()
        self.prompt = system_prompt
        self.model_name = model_name
        self.messages: List[Any] = [
            {
                "role": "system",
                "content": system_prompt,
            },
            {
                "role": "assistant",
                "content": initial_message,
            },
        ]

    def respond(self, human_input: str):
        self.messages.append(
            {
                "role": "user",
                "content": human_input,
            }
        )
        response = self.client.chat.completions.create(
            model=self.model_name,
            messages=self.messages,
        )
        content = response.choices[0].message.content
        self.messages.append(
            {
                "role": "system",
                "content": content,
            }
        )
        return content



import json
import logging
import sys

from loguru import logger
from loguru._handler import Handler

from vocode import get_serialized_ctx_wrappers


def _patched_serialize_record(text: str, record: dict) -> str:
    """
    This function takes a text string and a record dictionary as input and returns a serialized
    string representation of the record.

    The record dictionary is expected to contain various keys related to logging information such as
    'level', 'time', 'elapsed', 'exception', 'extra', 'file', 'function', 'line', 'message',
    'module', 'name', 'process', 'thread'. Each key's value is processed and added to a new
    dictionary 'serializable'.

    If the 'exception' key in the record is not None, it is further processed to extract 'type',
    'value', and 'traceback' information.

    The 'serializable' dictionary is then converted to a JSON string using json.dumps. The 'default'
    parameter is set to str to convert any non-serializable types to string. The 'ensure_ascii'
    parameter is set to False so that the function can output non-ASCII characters as they are.

    The function finally returns the serialized string with a newline character appended at the end.

    Args:
        text (str): A text string. record (dict): A dictionary containing logging information.

    Returns:
        str: A serialized string representation of the record dictionary.
    """
    exception = record["exception"]

    if exception is not None:
        exception = {
            "type": None if exception.type is None else exception.type.__name__,
            "value": exception.value,
            "traceback": bool(exception.traceback),
        }

    serializable = {
        "severity": record["level"].name,
        "text": text,
        "timestamp": record["time"].timestamp(),
        "elapsed": {
            "repr": record["elapsed"],
            "seconds": record["elapsed"].total_seconds(),
        },
        "exception": exception,
        "ctx": get_serialized_ctx_wrappers(),
        "extra": record["extra"],
        "file": {"name": record["file"].name, "path": record["file"].path},
        "function": record["function"],
        "level": {
            "icon": record["level"].icon,
            "name": record["level"].name,
            "no": record["level"].no,
        },
        "line": record["line"],
        "message": record["message"],
        "module": record["module"],
        "name": record["name"],
        "process": {"id": record["process"].id, "name": record["process"].name},
        "thread": {"id": record["thread"].id, "name": record["thread"].name},
        "time": {"repr": record["time"], "timestamp": record["time"].timestamp()},
    }

    return json.dumps(serializable, default=str, ensure_ascii=False) + "\n"


Handler._serialize_record = staticmethod(_patched_serialize_record)  # type: ignore


class InterceptHandler(logging.Handler):
    """
    Default handler from examples in loguru documentation.

    This handler intercepts all log requests and
    passes them to loguru.

    For more info see:
    https://loguru.readthedocs.io/en/stable/overview.html#entirely-compatible-with-standard-logging
    """

    def emit(self, record: logging.LogRecord) -> None:  # pragma: no cover
        """
        Propagates logs to loguru.

        :param record: record to log.
        """
        try:
            level: str | int = logger.level(record.levelname).name
        except ValueError:
            level = record.levelno

        # Find caller from where originated the logged message
        frame, depth = logging.currentframe(), 2
        while (
            frame.f_code.co_filename == logging.__file__
            or frame.f_code.co_filename == __file__
            or "sentry_sdk/integrations" in frame.f_code.co_filename
        ):
            frame = frame.f_back  # type: ignore
            depth += 1
        logger.opt(depth=depth, exception=record.exc_info).log(
            level,
            record.getMessage(),
        )


def configure_intercepter() -> None:
    """
    Configures the logging system to intercept log messages.

    This function sets up an InterceptHandler instance as the main handler for the root logger.
    It sets the logging level to INFO, meaning that all messages with severity INFO and above will be handled.

    It then iterates over all the loggers in the logging system. If a logger's name starts with "uvicorn.",
    it removes all handlers from that logger. This is done to prevent uvicorn's default logging configuration
    from interfering with our custom configuration.

    Finally, it sets the InterceptHandler instance as the sole handler for the "uvicorn" and "uvicorn.access" loggers.
    This ensures that all log messages from uvicorn and its access logger are intercepted by our custom handler.
    """
    intercept_handler = InterceptHandler()
    logging.basicConfig(handlers=[intercept_handler], level=logging.INFO)

    for logger_name in logging.root.manager.loggerDict:
        if logger_name.startswith("uvicorn."):
            logging.getLogger(logger_name).handlers = []

    logging.getLogger("uvicorn").handlers = [intercept_handler]
    logging.getLogger("uvicorn.access").handlers = [intercept_handler]


def configure_pretty_logging() -> None:
    """
    Configures the logging system to output pretty logs.

    This function enables the 'vocode' logger, sets up an intercept handler to
    capture logs from the standard logging module, removes all existing handlers
    from the 'loguru' logger, and adds a new handler that outputs to stdout with
    pretty formatting (colored, not serialized, no backtrace or diagnosis information).
    """
    logger.enable("vocode")

    configure_intercepter()

    logger.remove()
    logger.add(
        sys.stdout,
        level=logging.DEBUG,
        backtrace=False,
        diagnose=False,
        serialize=False,
        colorize=True,
    )


def configure_json_logging() -> None:
    """
    Configures the logging system to output logs in JSON format.

    This function enables the 'vocode' logger, sets up an intercept handler to
    capture logs from the standard logging module, removes all existing handlers
    from the 'loguru' logger, and adds a new handler that outputs to stdout with
    JSON formatting (serialized, no backtrace or diagnosis information).
    """
    logger.enable("vocode")

    configure_intercepter()

    logger.remove()
    logger.add(
        sys.stdout,
        format="{message}",
        level=logging.DEBUG,
        backtrace=False,
        diagnose=False,
        serialize=True,
    )



from typing import TYPE_CHECKING, Optional

from vocode.streaming.models.transcriber import EndpointingConfig
from vocode.streaming.synthesizer.input_streaming_synthesizer import InputStreamingSynthesizer
from vocode.streaming.telephony.client.twilio_client import TwilioClient
from vocode.streaming.telephony.client.vonage_client import VonageClient
from vocode.streaming.utils.redis_conversation_message_queue import RedisConversationMessageQueue

if TYPE_CHECKING:
    from vocode.streaming.streaming_conversation import StreamingConversation
    from vocode.streaming.telephony.conversation.abstract_phone_conversation import (
        AbstractPhoneConversation,
    )
    from vocode.streaming.telephony.conversation.twilio_phone_conversation import (
        TwilioPhoneConversation,
    )
    from vocode.streaming.telephony.conversation.vonage_phone_conversation import (
        VonagePhoneConversation,
    )


# TODO: make this a proper ABC
class AbstractConversationStateManager:
    @property
    def logger(self):
        raise NotImplementedError

    @property
    def transcript(self):
        raise NotImplementedError

    def get_transcriber_endpointing_config(self) -> Optional[EndpointingConfig]:
        raise NotImplementedError

    def set_transcriber_endpointing_config(self, endpointing_config: EndpointingConfig):
        raise NotImplementedError

    def disable_synthesis(self):
        raise NotImplementedError

    def enable_synthesis(self):
        raise NotImplementedError

    def mute_agent(self):
        raise NotImplementedError

    def unmute_agent(self):
        raise NotImplementedError

    def using_input_streaming_synthesizer(self):
        raise NotImplementedError

    async def terminate_conversation(self):
        raise NotImplementedError

    def get_conversation_id(self):
        raise NotImplementedError


class AbstractPhoneConversationStateManager(AbstractConversationStateManager):
    def get_config_manager(self):
        raise NotImplementedError

    def get_to_phone(self):
        raise NotImplementedError

    def get_from_phone(self):
        raise NotImplementedError


class ConversationStateManager(AbstractConversationStateManager):
    def __init__(self, conversation: "StreamingConversation"):
        self._conversation = conversation
        if not hasattr(self, "redis_message_queue"):
            self.redis_message_queue = RedisConversationMessageQueue()

    @property
    def transcript(self):
        return self._conversation.transcript

    def get_transcriber_endpointing_config(self) -> Optional[EndpointingConfig]:
        return self._conversation.transcriber.get_transcriber_config().endpointing_config

    def set_transcriber_endpointing_config(self, endpointing_config: EndpointingConfig):
        assert self.get_transcriber_endpointing_config() is not None
        self._conversation.transcriber.get_transcriber_config().endpointing_config = (
            endpointing_config
        )

    def disable_synthesis(self):
        self._conversation.synthesis_enabled = False

    def enable_synthesis(self):
        self._conversation.synthesis_enabled = True

    def mute_agent(self):
        self._conversation.agent.is_muted = True

    def unmute_agent(self):
        self._conversation.agent.is_muted = False

    def using_input_streaming_synthesizer(self):
        return isinstance(
            self._conversation.synthesizer,
            InputStreamingSynthesizer,
        )

    async def terminate_conversation(self):
        self._conversation.mark_terminated()

    def set_call_check_for_idle_paused(self, value: bool):
        if not self._conversation:
            return
        self._conversation.set_check_for_idle_paused(value)

    def get_conversation_id(self):
        return self._conversation.id


class PhoneConversationStateManager(
    AbstractPhoneConversationStateManager, ConversationStateManager
):
    def __init__(self, conversation: "AbstractPhoneConversation"):
        ConversationStateManager.__init__(self, conversation)
        self._phone_conversation = conversation

    def get_config_manager(self):
        return self._phone_conversation.config_manager

    def get_to_phone(self):
        return self._phone_conversation.to_phone

    def get_from_phone(self):
        return self._phone_conversation.from_phone

    def get_direction(self):
        return self._phone_conversation.direction


class VonagePhoneConversationStateManager(PhoneConversationStateManager):
    def __init__(self, conversation: "VonagePhoneConversation"):
        super().__init__(conversation=conversation)
        self._vonage_phone_conversation = conversation

    def create_vonage_client(self):
        return VonageClient(
            base_url=self._vonage_phone_conversation.base_url,
            maybe_vonage_config=self._vonage_phone_conversation.vonage_config,
        )


class TwilioPhoneConversationStateManager(PhoneConversationStateManager):
    def __init__(self, conversation: "TwilioPhoneConversation"):
        super().__init__(conversation=conversation)
        self._twilio_phone_conversation = conversation

    def get_twilio_config(self):
        return self._twilio_phone_conversation.twilio_config

    def create_twilio_client(self):
        return TwilioClient(
            base_url=self._twilio_phone_conversation.base_url,
            maybe_twilio_config=self.get_twilio_config(),
        )



import io

import aiohttp
from pydub import AudioSegment

from vocode.streaming.models.message import BaseMessage
from vocode.streaming.models.synthesizer import StreamElementsSynthesizerConfig
from vocode.streaming.synthesizer.base_synthesizer import BaseSynthesizer, SynthesisResult


class StreamElementsSynthesizer(BaseSynthesizer[StreamElementsSynthesizerConfig]):
    TTS_ENDPOINT = "https://api.streamelements.com/kappa/v2/speech"

    def __init__(
        self,
        synthesizer_config: StreamElementsSynthesizerConfig,
    ):
        super().__init__(synthesizer_config)
        self.voice = synthesizer_config.voice

    async def create_speech(
        self,
        message: BaseMessage,
        chunk_size: int,
        is_first_text_chunk: bool = False,
        is_sole_text_chunk: bool = False,
    ) -> SynthesisResult:
        url_params = {
            "voice": self.voice,
            "text": message.text,
        }
        async with self.async_requestor.get_session().get(
            self.TTS_ENDPOINT,
            params=url_params,
            timeout=aiohttp.ClientTimeout(total=15),
        ) as response:
            read_response = await response.read()

            # TODO: probably needs to be in a thread
            audio_segment: AudioSegment = AudioSegment.from_mp3(
                io.BytesIO(read_response)  # type: ignore
            )
            output_bytes_io = io.BytesIO()
            audio_segment.export(output_bytes_io, format="wav")  # type: ignore

            result = self.create_synthesis_result_from_wav(
                synthesizer_config=self.synthesizer_config,
                file=output_bytes_io,
                message=message,
                chunk_size=chunk_size,
            )

            return result



import asyncio
import json
import time
from typing import Optional

import websockets
from loguru import logger
from websockets.client import WebSocketClientProtocol

from vocode import getenv
from vocode.streaming.models.transcriber import (
    RevAITranscriberConfig,
    TimeEndpointingConfig,
    Transcription,
)
from vocode.streaming.transcriber.base_transcriber import BaseAsyncTranscriber

NUM_RESTARTS = 5


def getSeconds():
    return time.time()


class RevAITranscriber(BaseAsyncTranscriber[RevAITranscriberConfig]):
    def __init__(
        self,
        transcriber_config: RevAITranscriberConfig,
        api_key: Optional[str] = None,
    ):
        super().__init__(transcriber_config)
        self.api_key = api_key or getenv("REV_AI_API_KEY")
        if not self.api_key:
            raise Exception(
                "Please set REV_AI_API_KEY environment variable or pass it as a parameter"
            )
        self.closed = False
        self.is_ready = True
        self.last_signal_seconds = 0

    async def ready(self):
        return self.is_ready

    def get_rev_ai_url(self):
        codec = "audio/x-raw"
        layout = "interleaved"
        rate = self.get_transcriber_config().sampling_rate
        audio_format = "S16LE"
        channels = 1

        content_type = (
            f"{codec};layout={layout};rate={rate};format={audio_format};channels={channels}"
        )

        url_params_dict = {
            "access_token": self.api_key,
            "content_type": content_type,
        }

        url_params_arr = [f"{key}={value}" for (key, value) in url_params_dict.items()]
        url = "wss://api.rev.ai/speechtotext/v1/stream?" + "&".join(url_params_arr)
        return url

    async def _run_loop(self):
        restarts = 0
        while not self.closed and restarts < NUM_RESTARTS:
            await self.process()
            restarts += 1
            logger.debug(f"Rev AI connection died, restarting, num_restarts: {restarts}")

    async def process(self):
        async with websockets.connect(self.get_rev_ai_url()) as ws:

            async def sender(ws: WebSocketClientProtocol):
                while not self.closed:
                    try:
                        data = await asyncio.wait_for(self.input_queue.get(), 5)
                    except asyncio.exceptions.TimeoutError:
                        break
                    await ws.send(data)
                await ws.close()
                logger.debug("Terminating Rev.AI transcriber sender")

            async def receiver(ws: WebSocketClientProtocol):
                buffer = ""

                while not self.closed:
                    try:
                        msg = await ws.recv()
                    except Exception as e:
                        logger.debug(f"Got error {e} in Rev.AI receiver")
                        break
                    data = json.loads(msg)

                    if data["type"] == "connected":
                        continue

                    is_done = data["type"] == "final"
                    if (
                        (len(buffer) > 0)
                        and (self.transcriber_config.endpointing_config)
                        and isinstance(
                            self.transcriber_config.endpointing_config,
                            TimeEndpointingConfig,
                        )
                        and (
                            getSeconds()
                            > self.last_signal_seconds
                            + self.transcriber_config.endpointing_config.time_cutoff_seconds
                        )
                    ):
                        is_done = True

                    new_text = "".join([e["value"] for e in data["elements"]])
                    if len(new_text) > len(buffer):
                        self.last_signal_seconds = getSeconds()
                    buffer = new_text

                    confidence = 1.0
                    if is_done:
                        self.output_queue.put_nowait(
                            Transcription(message=buffer, confidence=confidence, is_final=True)
                        )
                        buffer = ""
                    else:
                        self.output_queue.put_nowait(
                            Transcription(
                                message=buffer,
                                confidence=confidence,
                                is_final=False,
                            )
                        )

                logger.debug("Terminating Rev.AI transcriber receiver")

            await asyncio.gather(sender(ws), receiver(ws))

    def terminate(self):
        terminate_msg = json.dumps({"type": "CloseStream"})
        self.input_queue.put_nowait(terminate_msg)
        self.closed = True



import asyncio
import audioop
import random
import secrets
import wave
from string import ascii_letters, digits
from typing import Any, AsyncGenerator, AsyncIterator, Callable, List, Tuple, TypeVar

from vocode.streaming.models.audio import AudioEncoding

custom_alphabet = ascii_letters + digits + ".-_"

ChoiceType = TypeVar("ChoiceType")


def create_loop_in_thread(loop: asyncio.AbstractEventLoop, long_running_task=None):
    asyncio.set_event_loop(loop)
    if long_running_task:
        loop.run_until_complete(long_running_task)
    else:
        loop.run_forever()


def convert_linear_audio(
    raw_wav: bytes,
    input_sample_rate=24000,
    output_sample_rate=8000,
    output_encoding=AudioEncoding.LINEAR16,
    output_sample_width=2,
):
    # downsample
    if input_sample_rate != output_sample_rate:
        raw_wav, _ = audioop.ratecv(raw_wav, 2, 1, input_sample_rate, output_sample_rate, None)

    if output_encoding == AudioEncoding.LINEAR16:
        return raw_wav
    elif output_encoding == AudioEncoding.MULAW:
        return audioop.lin2ulaw(raw_wav, output_sample_width)


def convert_wav(
    file: Any,
    output_sample_rate=8000,
    output_encoding=AudioEncoding.LINEAR16,
):
    with wave.open(file, "rb") as wav:
        raw_wav = wav.readframes(wav.getnframes())
        return convert_linear_audio(
            raw_wav,
            input_sample_rate=wav.getframerate(),
            output_sample_rate=output_sample_rate,
            output_encoding=output_encoding,
            output_sample_width=wav.getsampwidth(),
        )


def get_chunk_size_per_second(audio_encoding: AudioEncoding, sampling_rate: int) -> int:
    if audio_encoding == AudioEncoding.LINEAR16:
        return sampling_rate * 2
    elif audio_encoding == AudioEncoding.MULAW:
        return sampling_rate
    else:
        raise Exception("Unsupported audio encoding")


def create_conversation_id() -> str:
    return secrets.token_urlsafe(16)


def create_utterance_id() -> str:
    return secrets.token_hex(16)


def remove_non_letters_digits(text):
    return "".join(i for i in text if i in custom_alphabet)


def unrepeating_randomizer(l: List[ChoiceType]) -> Callable[[], ChoiceType]:
    last_choice = None

    def next() -> ChoiceType:
        nonlocal last_choice
        choice = random.choice(l)
        while choice == last_choice:
            choice = random.choice(l)
        last_choice = choice
        return choice

    return next


AsyncIteratorGenericType = TypeVar("AsyncIteratorGenericType")


async def generate_with_is_last(
    async_gen: AsyncGenerator[AsyncIteratorGenericType, None],
) -> AsyncGenerator[Tuple[AsyncIteratorGenericType, bool], None]:
    async_iter = async_gen.__aiter__()
    try:
        next_item = await async_iter.__anext__()
    except StopAsyncIteration:
        assert False, "Cannot generate with is_last from an empty async generator"
    while True:
        try:
            item = await async_iter.__anext__()
            yield next_item, False
            next_item = item
        except StopAsyncIteration:
            yield next_item, True
            break


async def generate_from_async_iter_with_lookahead(
    async_iter: AsyncIterator[AsyncIteratorGenericType],
    lookahead: int,
) -> AsyncGenerator[List[AsyncIteratorGenericType], None]:
    """Yield sliding window lists of length `lookahead + 1` from an async iterator.

    If the length of async iterator < lookahead + 1, then it should just yield the whole
        async iterator as a list.
    """
    assert lookahead > 0
    buffer = []

    stream_length = 0
    while True:
        try:
            next_item = await async_iter.__anext__()
            stream_length += 1
            buffer.append(next_item)
            if len(buffer) == lookahead + 1:
                yield buffer
                buffer = buffer[1:]
        except StopAsyncIteration:
            if buffer and stream_length <= lookahead:
                yield buffer
            return



from typing import List


def split_text(string_to_split: str, max_text_length: int) -> List[str]:
    # Base case: if the string_to_split is less than or equal to max_text_length characters, return it as a single element array
    if len(string_to_split) <= max_text_length:
        return [string_to_split.strip()]

    # Recursive case: find the index of the last sentence ender in the first max_text_length characters of the string_to_split
    sentence_enders = [".", "!", "?"]
    index = -1
    for ender in sentence_enders:
        i = string_to_split[:max_text_length].rfind(ender)
        if i > index:
            index = i

    # If there is a sentence ender, split the string_to_split at that index plus one and strip any spaces from both parts
    if index != -1:
        first_part = string_to_split[: index + 1].strip()
        second_part = string_to_split[index + 1 :].strip()

    # If there is no sentence ender, find the index of the last comma in the first max_text_length characters of the string_to_split
    else:
        index = string_to_split[:max_text_length].rfind(",")
        # If there is a comma, split the string_to_split at that index plus one and strip any spaces from both parts
        if index != -1:
            first_part = string_to_split[: index + 1].strip()
            second_part = string_to_split[index + 1 :].strip()
        # If there is no comma, find the index of the last space in the first max_text_length characters of the string_to_split
        else:
            index = string_to_split[:max_text_length].rfind(" ")
            # If there is a space, split the string_to_split at that index and strip any spaces from both parts
            if index != -1:
                first_part = string_to_split[:index].strip()
                second_part = string_to_split[index:].strip()

            # If there is no space, split the string_to_split at max_text_length characters and strip any spaces from both parts
            else:
                first_part = string_to_split[:max_text_length].strip()
                second_part = string_to_split[max_text_length:].strip()

    # Append the first part to the result array
    result = [first_part]

    # Call the function recursively on the remaining part of the string_to_split and extend the result array with it, unless it is empty
    if second_part != "":
        result.extend(split_text(string_to_split=second_part, max_text_length=max_text_length))

    # Return the result array
    return result



from abc import ABC, abstractmethod

from vocode.streaming.agent.base_agent import BaseAgent
from vocode.streaming.models.agent import AgentConfig


class AbstractAgentFactory(ABC):
    @abstractmethod
    def create_agent(self, agent_config: AgentConfig) -> BaseAgent:
        pass



from __future__ import annotations

import asyncio
import threading
from typing import Any, Generic, Optional, TypeVar

import janus
from loguru import logger

from vocode.streaming.utils.create_task import asyncio_create_task_with_done_error_log

WorkerInputType = TypeVar("WorkerInputType")


class AsyncWorker(Generic[WorkerInputType]):
    def __init__(
        self,
        input_queue: asyncio.Queue,
        output_queue: asyncio.Queue = asyncio.Queue(),
    ) -> None:
        self.worker_task: Optional[asyncio.Task] = None
        self.input_queue = input_queue
        self.output_queue = output_queue

    def start(self) -> asyncio.Task:
        self.worker_task = asyncio_create_task_with_done_error_log(
            self._run_loop(),
        )
        if not self.worker_task:
            raise Exception("Worker task not created")
        return self.worker_task

    def consume_nonblocking(self, item: WorkerInputType):
        self.input_queue.put_nowait(item)

    def produce_nonblocking(self, item):
        self.output_queue.put_nowait(item)

    async def _run_loop(self):
        raise NotImplementedError

    def terminate(self):
        if self.worker_task:
            return self.worker_task.cancel()

        return False


class ThreadAsyncWorker(AsyncWorker[WorkerInputType]):
    def __init__(
        self,
        input_queue: asyncio.Queue[WorkerInputType],
        output_queue: asyncio.Queue = asyncio.Queue(),
    ) -> None:
        super().__init__(input_queue, output_queue)
        self.worker_thread: Optional[threading.Thread] = None
        self.input_janus_queue: janus.Queue[WorkerInputType] = janus.Queue()
        self.output_janus_queue: janus.Queue = janus.Queue()

    def start(self) -> asyncio.Task:
        self.worker_thread = threading.Thread(target=self._run_loop)
        self.worker_thread.start()
        self.worker_task = asyncio_create_task_with_done_error_log(
            self.run_thread_forwarding(),
        )
        if not self.worker_task:
            raise Exception("Worker task not created")
        return self.worker_task

    async def run_thread_forwarding(self):
        try:
            await asyncio.gather(
                self._forward_to_thread(),
                self._forward_from_thead(),
            )
        except asyncio.CancelledError:
            return

    async def _forward_to_thread(self):
        while True:
            item = await self.input_queue.get()
            self.input_janus_queue.async_q.put_nowait(item)

    async def _forward_from_thead(self):
        while True:
            item = await self.output_janus_queue.async_q.get()
            self.output_queue.put_nowait(item)

    def _run_loop(self):
        raise NotImplementedError

    def terminate(self):
        return super().terminate()


class AsyncQueueWorker(AsyncWorker):
    async def _run_loop(self):
        while True:
            try:
                item = await self.input_queue.get()
                await self.process(item)
            except asyncio.CancelledError:
                return
            except Exception:
                logger.exception("AsyncQueueWorker", exc_info=True)

    async def process(self, item):
        """
        Publish results onto output queue.
        Calls to async function / task should be able to handle asyncio.CancelledError gracefully and not re-raise it
        """
        raise NotImplementedError


Payload = TypeVar("Payload")


class InterruptibleEvent(Generic[Payload]):
    def __init__(
        self,
        payload: Payload,
        is_interruptible: bool = True,
        interruption_event: Optional[threading.Event] = None,
    ):
        self.interruption_event = interruption_event or threading.Event()
        self.is_interruptible = is_interruptible
        self.payload = payload

    def interrupt(self) -> bool:
        """
        Returns True if the event was interruptible and is now interrupted.
        """
        if not self.is_interruptible:
            return False
        self.interruption_event.set()
        return True

    def is_interrupted(self):
        return self.is_interruptible and self.interruption_event.is_set()


class InterruptibleAgentResponseEvent(InterruptibleEvent[Payload]):
    def __init__(
        self,
        payload: Payload,
        agent_response_tracker: asyncio.Event,
        is_interruptible: bool = True,
        interruption_event: Optional[threading.Event] = None,
    ):
        super().__init__(payload, is_interruptible, interruption_event)
        self.agent_response_tracker = agent_response_tracker

    def interrupt(self) -> bool:
        self.agent_response_tracker.set()
        return super().interrupt()


class InterruptibleEventFactory:
    def create_interruptible_event(
        self, payload: Any, is_interruptible: bool = True
    ) -> InterruptibleEvent:
        return InterruptibleEvent(payload, is_interruptible=is_interruptible)

    def create_interruptible_agent_response_event(
        self,
        payload: Any,
        is_interruptible: bool = True,
        agent_response_tracker: Optional[asyncio.Event] = None,
    ) -> InterruptibleAgentResponseEvent:
        return InterruptibleAgentResponseEvent(
            payload,
            is_interruptible=is_interruptible,
            agent_response_tracker=agent_response_tracker or asyncio.Event(),
        )


InterruptibleEventType = TypeVar("InterruptibleEventType", bound=InterruptibleEvent)


class InterruptibleWorker(AsyncWorker[InterruptibleEventType]):
    def __init__(
        self,
        input_queue: asyncio.Queue[InterruptibleEventType],
        output_queue: asyncio.Queue = asyncio.Queue(),
        interruptible_event_factory: InterruptibleEventFactory = InterruptibleEventFactory(),
        max_concurrency=2,
    ) -> None:
        super().__init__(input_queue, output_queue)
        self.input_queue = input_queue
        self.max_concurrency = max_concurrency
        self.interruptible_event_factory = interruptible_event_factory
        self.current_task = None
        self.interruptible_event = None

    def produce_interruptible_event_nonblocking(self, item: Any, is_interruptible: bool = True):
        interruptible_event = self.interruptible_event_factory.create_interruptible_event(
            item, is_interruptible=is_interruptible
        )
        return super().produce_nonblocking(interruptible_event)

    def produce_interruptible_agent_response_event_nonblocking(
        self,
        item: Any,
        is_interruptible: bool = True,
        agent_response_tracker: Optional[asyncio.Event] = None,
    ):
        interruptible_utterance_event = (
            self.interruptible_event_factory.create_interruptible_agent_response_event(
                item,
                is_interruptible=is_interruptible,
                agent_response_tracker=agent_response_tracker or asyncio.Event(),
            )
        )
        return super().produce_nonblocking(interruptible_utterance_event)

    async def _run_loop(self):
        # TODO Implement concurrency with max_nb_of_thread
        while True:
            try:
                item = await self.input_queue.get()
            except asyncio.CancelledError:
                return

            if item.is_interrupted():
                continue
            self.interruptible_event = item
            self.current_task = asyncio_create_task_with_done_error_log(
                self.process(item),
                reraise_cancelled=True,
            )

            try:
                await self.current_task
            except asyncio.CancelledError:
                return
            except Exception:
                logger.exception("InterruptibleWorker", exc_info=True)
            self.interruptible_event.is_interruptible = False
            self.current_task = None

    async def process(self, item: InterruptibleEventType):
        """
        Publish results onto output queue.
        Calls to async function / task should be able to handle asyncio.CancelledError gracefully:
        """
        raise NotImplementedError

    def cancel_current_task(self):
        """Free up the resources. That's useful so implementors do not have to implement this but:
        - threads tasks won't be able to be interrupted. Hopefully not too much of a big deal
            Threads will also get a reference to the interruptible event
        - asyncio tasks will still have to handle CancelledError and clean up resources
        """
        if (
            self.current_task
            and not self.current_task.done()
            and self.interruptible_event.is_interruptible
        ):
            return self.current_task.cancel()

        return False


class InterruptibleAgentResponseWorker(InterruptibleWorker[InterruptibleAgentResponseEvent]):
    pass



from copy import deepcopy
from typing import Any, AsyncGenerator, Dict, List, Optional, Union

from loguru import logger
from openai.types.chat.chat_completion_chunk import ChatCompletionChunk

from vocode.streaming.agent.token_utils import (
    get_chat_gpt_max_tokens,
    num_tokens_from_functions,
    num_tokens_from_messages,
)
from vocode.streaming.models.actions import FunctionFragment, PhraseBasedActionTrigger
from vocode.streaming.models.agent import LLM_AGENT_DEFAULT_MAX_TOKENS
from vocode.streaming.models.events import Sender
from vocode.streaming.models.transcript import (
    ActionFinish,
    ActionStart,
    ConferenceEvent,
    EventLog,
    Message,
    Transcript,
)


def vector_db_result_to_openai_chat_message(vector_db_result):
    return {"role": "user", "content": vector_db_result}


def is_phrase_based_action_event_log(event_log: EventLog) -> bool:
    return (
        (isinstance(event_log, ActionStart) or isinstance(event_log, ActionFinish))
        and event_log.action_input is not None
        and event_log.action_input.action_config is not None
        and isinstance(
            event_log.action_input.action_config.action_trigger, PhraseBasedActionTrigger
        )
    )


def get_openai_chat_messages_from_transcript(
    merged_event_logs: List[EventLog],
    prompt_preamble: str,
) -> List[dict]:
    chat_messages = [{"role": "system", "content": prompt_preamble}]
    for event_log in merged_event_logs:
        if isinstance(event_log, Message):
            if len(event_log.text.strip()) == 0:
                continue
            else:
                chat_messages.append(
                    {
                        "role": ("assistant" if event_log.sender == Sender.BOT else "user"),
                        "content": event_log.to_string(include_sender=False),
                    },
                )
        elif isinstance(event_log, ActionStart):
            action_message: Dict[str, Any]
            if is_phrase_based_action_event_log(event_log=event_log):
                pass
            else:
                action_message = {
                    "role": "assistant",
                    "content": None,
                    "function_call": {
                        "name": event_log.action_type,
                        "arguments": event_log.action_input.params.json(),
                    },
                }
                chat_messages.append(action_message)
        elif isinstance(event_log, ActionFinish):
            action_message = {
                "role": "function",
                "name": event_log.action_type,
                "content": event_log.to_string(include_header=False),
            }
            chat_messages.append(action_message)
        elif isinstance(event_log, ConferenceEvent):
            chat_messages.append(
                {"role": "user", "content": event_log.to_string(include_sender=False)},
            )
    return chat_messages


def merge_event_logs(event_logs: List[EventLog]) -> List[EventLog]:
    """Returns a new list of event logs where consecutive bot messages are merged."""
    new_event_logs: List[EventLog] = []
    idx = 0
    while idx < len(event_logs):
        bot_messages_buffer: List[Message] = []
        current_log = event_logs[idx]
        while isinstance(current_log, Message) and current_log.sender == Sender.BOT:
            bot_messages_buffer.append(current_log)
            idx += 1
            try:
                current_log = event_logs[idx]
            except IndexError:
                break
        if bot_messages_buffer:
            merged_bot_message = deepcopy(bot_messages_buffer[-1])
            merged_bot_message.text = " ".join(event_log.text for event_log in bot_messages_buffer)
            new_event_logs.append(merged_bot_message)
        else:
            new_event_logs.append(current_log)
            idx += 1

    return new_event_logs


def format_openai_chat_messages_from_transcript(
    transcript: Transcript,
    model_name: str,
    functions: Optional[List[Dict]],
    prompt_preamble: str,
) -> List[dict]:
    # merge consecutive bot messages
    merged_event_logs: List[EventLog] = merge_event_logs(event_logs=transcript.event_logs)

    chat_messages: List[Dict[str, Optional[Any]]]
    chat_messages = get_openai_chat_messages_from_transcript(
        merged_event_logs=merged_event_logs,
        prompt_preamble=prompt_preamble,
    )

    context_size = num_tokens_from_messages(
        messages=chat_messages,
        model=model_name,
    ) + num_tokens_from_functions(functions=functions, model=model_name)

    num_removed_messages = 0
    while (
        context_size > get_chat_gpt_max_tokens(model_name) - LLM_AGENT_DEFAULT_MAX_TOKENS - 50
    ):  # context limit includes the max tokens, and 50 for safety
        if len(chat_messages) <= 1:
            logger.error(f"Prompt is too long to fit in context window, num tokens {context_size}")
            break
        num_removed_messages += 1
        chat_messages.pop(1)
        context_size = num_tokens_from_messages(
            messages=chat_messages,
            model=model_name,
        ) + num_tokens_from_functions(functions=functions, model=model_name)

    if num_removed_messages > 0:
        logger.info(
            "Removed %d messages from prompt to satisfy context limit",
            num_removed_messages,
        )

    return chat_messages


async def openai_get_tokens(
    gen: AsyncGenerator[ChatCompletionChunk, None],
) -> AsyncGenerator[Union[str, FunctionFragment], None]:
    async for event in gen:
        choices = event.choices
        if len(choices) == 0:
            continue
        choice = choices[0]
        if choice.finish_reason:
            if choice.finish_reason == "content_filter":
                logger.warning(
                    "Detected content filter.",
                    extra={"chat_completion_chunk": event.model_dump()},
                )
            break
        delta = choice.delta
        if delta.content is not None:
            token = delta.content
            yield token
        elif delta.function_call is not None:
            yield FunctionFragment(
                name=(delta.function_call.name if delta.function_call.name is not None else ""),
                arguments=(
                    delta.function_call.arguments
                    if delta.function_call.arguments is not None
                    else ""
                ),
            )



import asyncio
import audioop
import json
from typing import Optional
from urllib.parse import urlencode

import numpy as np
import websockets
from loguru import logger

from vocode import getenv
from vocode.streaming.models.audio import AudioEncoding
from vocode.streaming.models.transcriber import (
    AssemblyAITranscriberConfig,
    PunctuationEndpointingConfig,
    TimeEndpointingConfig,
    Transcription,
)
from vocode.streaming.models.websocket import AudioMessage
from vocode.streaming.transcriber.base_transcriber import BaseAsyncTranscriber

ASSEMBLY_AI_URL = "wss://api.assemblyai.com/v2/realtime/ws"


class AssemblyAITranscriber(BaseAsyncTranscriber[AssemblyAITranscriberConfig]):
    def __init__(
        self,
        transcriber_config: AssemblyAITranscriberConfig,
        api_key: Optional[str] = None,
    ):
        super().__init__(transcriber_config)
        self.api_key = api_key or getenv("ASSEMBLY_AI_API_KEY")
        if not self.api_key:
            raise Exception(
                "Please set ASSEMBLY_AI_API_KEY environment variable or pass it as a parameter"
            )
        self._ended = False
        self.buffer = bytearray()
        self.audio_cursor = 0

        if isinstance(
            self.transcriber_config.endpointing_config,
            (TimeEndpointingConfig, PunctuationEndpointingConfig),
        ):
            self.transcriber_config.end_utterance_silence_threshold_milliseconds = int(
                self.transcriber_config.endpointing_config.time_cutoff_seconds * 1000
            )
        self.terminate_msg = json.dumps({"terminate_session": True})
        self.end_utterance_silence_threshold_msg = (
            None
            if self.transcriber_config.end_utterance_silence_threshold_milliseconds is None
            else json.dumps(
                {
                    "end_utterance_silence_threshold": self.transcriber_config.end_utterance_silence_threshold_milliseconds
                }
            )
        )

    async def ready(self):
        return True

    async def _run_loop(self):
        await self.process()

    def send_audio(self, chunk):
        if self.transcriber_config.audio_encoding == AudioEncoding.MULAW:
            sample_width = 1
            if isinstance(chunk, np.ndarray):
                chunk = chunk.astype(np.int16)
                chunk = chunk.tobytes()
            chunk = audioop.ulaw2lin(chunk, sample_width)

        self.buffer.extend(chunk)

        if (
            len(self.buffer) / (2 * self.transcriber_config.sampling_rate)
        ) >= self.transcriber_config.buffer_size_seconds:
            self.input_queue.put_nowait(self.buffer)
            self.buffer = bytearray()

    def terminate(self):
        self._ended = True
        super().terminate()

    def get_assembly_ai_url(self):
        url_params = {"sample_rate": self.transcriber_config.sampling_rate}
        if self.transcriber_config.word_boost:
            url_params.update({"word_boost": json.dumps(self.transcriber_config.word_boost)})
        return ASSEMBLY_AI_URL + f"?{urlencode(url_params)}"

    async def process(self):
        self.audio_cursor = 0
        URL = self.get_assembly_ai_url()

        async with websockets.connect(
            URL,
            extra_headers=(("Authorization", self.api_key),),
            ping_interval=5,
            ping_timeout=20,
        ) as ws:
            await asyncio.sleep(0.1)

            if self.end_utterance_silence_threshold_msg:
                await ws.send(self.end_utterance_silence_threshold_msg)

            async def sender(ws):  # sends audio to websocket
                while not self._ended:
                    try:
                        data = await asyncio.wait_for(self.input_queue.get(), 5)
                    except asyncio.exceptions.TimeoutError:
                        break
                    num_channels = 1
                    sample_width = 2
                    self.audio_cursor += len(data) / (
                        self.transcriber_config.sampling_rate * num_channels * sample_width
                    )
                    await ws.send(json.dumps({"audio_data": AudioMessage.from_bytes(data).data}))
                await ws.send(self.terminate_msg)
                logger.debug("Terminating AssemblyAI transcriber sender")

            async def receiver(ws):
                while not self._ended:
                    try:
                        result_str = await ws.recv()
                        data = json.loads(result_str)
                        if "error" in data and data["error"]:
                            raise Exception(data["error"])
                    except websockets.exceptions.ConnectionClosedError as e:
                        logger.debug(e)
                        break

                    data = json.loads(result_str)
                    is_final = "message_type" in data and data["message_type"] == "FinalTranscript"

                    if "text" in data and data["text"]:
                        self.output_queue.put_nowait(
                            Transcription(
                                message=data["text"],
                                confidence=data["confidence"],
                                is_final=is_final,
                            )
                        )

            await asyncio.gather(sender(ws), receiver(ws))



import asyncio
import audioop
import io
import math
import os
import wave
from typing import Any, AsyncGenerator, Callable, Generic, List, Optional, Tuple, TypeVar, Union

import aiohttp
from loguru import logger
from nltk.tokenize import word_tokenize
from nltk.tokenize.treebank import TreebankWordDetokenizer
from sentry_sdk.tracing import Span as SentrySpan

from vocode.streaming.models.agent import FillerAudioConfig
from vocode.streaming.models.audio import AudioEncoding, SamplingRate
from vocode.streaming.models.message import BaseMessage, BotBackchannel, SilenceMessage
from vocode.streaming.models.synthesizer import SynthesizerConfig
from vocode.streaming.synthesizer.audio_cache import AudioCache
from vocode.streaming.synthesizer.miniaudio_worker import MiniaudioWorker
from vocode.streaming.telephony.constants import MULAW_SILENCE_BYTE, PCM_SILENCE_BYTE
from vocode.streaming.utils import convert_wav, get_chunk_size_per_second
from vocode.streaming.utils.async_requester import AsyncRequestor
from vocode.streaming.utils.create_task import asyncio_create_task_with_done_error_log

FILLER_PHRASES = [
    BaseMessage(text="Um..."),
    BaseMessage(text="Uh..."),
    BaseMessage(text="Uh-huh..."),
    BaseMessage(text="Mm-hmm..."),
    BaseMessage(text="Hmm..."),
    BaseMessage(text="Okay..."),
    BaseMessage(text="Right..."),
    BaseMessage(text="Let me see..."),
]
FILLER_AUDIO_PATH = os.path.join(os.path.dirname(__file__), "filler_audio")
TYPING_NOISE_PATH = "%s/typing-noise.wav" % FILLER_AUDIO_PATH


def encode_as_wav(chunk: bytes, synthesizer_config: SynthesizerConfig) -> bytes:
    output_bytes_io = io.BytesIO()
    in_memory_wav = wave.open(output_bytes_io, "wb")
    in_memory_wav.setnchannels(1)
    assert synthesizer_config.audio_encoding == AudioEncoding.LINEAR16
    in_memory_wav.setsampwidth(2)
    in_memory_wav.setframerate(synthesizer_config.sampling_rate)
    in_memory_wav.writeframes(chunk)
    output_bytes_io.seek(0)
    return output_bytes_io.read()


class SynthesisResult:
    class ChunkResult:
        def __init__(self, chunk: bytes, is_last_chunk: bool):
            self.chunk = chunk
            self.is_last_chunk = is_last_chunk

    def __init__(
        self,
        chunk_generator: AsyncGenerator[ChunkResult, None],
        get_message_up_to: Callable[[float], str],
        cached: bool = False,
        is_first: bool = False,
        synthesis_total_span: Optional[SentrySpan] = None,
        ttft_span: Optional[SentrySpan] = None,
    ):
        self.chunk_generator = chunk_generator
        self.get_message_up_to = get_message_up_to
        self.cached = cached
        self.is_first = is_first
        self.synthesis_total_span = synthesis_total_span
        self.ttft_span = ttft_span


class FillerAudio:
    def __init__(
        self,
        message: BaseMessage,
        audio_data: bytes,
        synthesizer_config: SynthesizerConfig,
        is_interruptible: bool = False,
        seconds_per_chunk: int = 1,
    ):
        self.message = message
        self.audio_data = audio_data
        self.synthesizer_config = synthesizer_config
        self.is_interruptible = is_interruptible
        self.seconds_per_chunk = seconds_per_chunk

    def create_synthesis_result(self) -> SynthesisResult:
        chunk_size = (
            get_chunk_size_per_second(
                self.synthesizer_config.audio_encoding,
                self.synthesizer_config.sampling_rate,
            )
            * self.seconds_per_chunk
        )

        async def chunk_generator(chunk_transform=lambda x: x):
            for i in range(0, len(self.audio_data), chunk_size):
                if i + chunk_size > len(self.audio_data):
                    yield SynthesisResult.ChunkResult(chunk_transform(self.audio_data[i:]), True)
                else:
                    yield SynthesisResult.ChunkResult(
                        chunk_transform(self.audio_data[i : i + chunk_size]), False
                    )

        if self.synthesizer_config.should_encode_as_wav:
            output_generator = chunk_generator(
                lambda chunk: encode_as_wav(chunk, self.synthesizer_config)
            )
        else:
            output_generator = chunk_generator()
        return SynthesisResult(output_generator, lambda seconds: self.message.text)


class CachedAudio:
    def __init__(
        self,
        message: BaseMessage,
        audio_data: bytes,
        synthesizer_config: SynthesizerConfig,
        trailing_silence_seconds: float = 0.0,
    ):
        self.message = message
        self.audio_data = audio_data
        self.synthesizer_config = synthesizer_config
        self.trailing_silence_seconds = trailing_silence_seconds

    def create_synthesis_result(self, chunk_size) -> SynthesisResult:
        async def chunk_generator():
            if isinstance(self.message, BotBackchannel):
                yield SynthesisResult.ChunkResult(
                    self.audio_data, self.trailing_silence_seconds == 0.0
                )
            else:
                for i in range(0, len(self.audio_data), chunk_size):
                    if i + chunk_size > len(self.audio_data):
                        yield SynthesisResult.ChunkResult(
                            self.audio_data[i:], self.trailing_silence_seconds == 0.0
                        )
                    else:
                        yield SynthesisResult.ChunkResult(
                            self.audio_data[i : i + chunk_size], False
                        )
            if self.trailing_silence_seconds > 0:
                silence_synthesis_result = self.create_silence_synthesis_result(chunk_size)
                async for chunk_result in silence_synthesis_result.chunk_generator:
                    yield chunk_result

        if isinstance(self.message, BotBackchannel):

            def get_message_up_to(seconds):
                return self.message.text

        else:

            def get_message_up_to(seconds):
                return BaseSynthesizer.get_message_cutoff_from_total_response_length(
                    self.synthesizer_config, self.message, seconds, len(self.audio_data)
                )

        return SynthesisResult(
            chunk_generator=chunk_generator(),
            get_message_up_to=get_message_up_to,
            cached=True,
        )

    def create_silence_synthesis_result(self, chunk_size) -> SynthesisResult:
        async def chunk_generator():
            size_of_silence = int(
                self.trailing_silence_seconds * self.synthesizer_config.sampling_rate
            )
            silence_byte: bytes
            if self.synthesizer_config.audio_encoding == AudioEncoding.LINEAR16:
                silence_byte = PCM_SILENCE_BYTE
                size_of_silence *= 2
            elif self.synthesizer_config.audio_encoding == AudioEncoding.MULAW:
                silence_byte = MULAW_SILENCE_BYTE

            for _ in range(
                0,
                size_of_silence,
                chunk_size,
            ):
                yield SynthesisResult.ChunkResult(silence_byte * chunk_size, False)
            yield SynthesisResult.ChunkResult(silence_byte * chunk_size, True)

        def get_message_up_to(seconds):
            return ""

        return SynthesisResult(
            chunk_generator(),
            get_message_up_to,
        )


class SilenceAudio(CachedAudio):
    def __init__(
        self,
        message: SilenceMessage,
        synthesizer_config: SynthesizerConfig,
    ):
        super().__init__(
            message,
            b"",
            synthesizer_config,
            trailing_silence_seconds=message.trailing_silence_seconds,
        )

    def create_synthesis_result(self, chunk_size) -> SynthesisResult:
        return self.create_silence_synthesis_result(chunk_size)


SynthesizerConfigType = TypeVar("SynthesizerConfigType", bound=SynthesizerConfig)


class BaseSynthesizer(Generic[SynthesizerConfigType]):
    def __init__(
        self,
        synthesizer_config: SynthesizerConfigType,
    ):
        self.synthesizer_config = synthesizer_config
        if synthesizer_config.audio_encoding == AudioEncoding.MULAW:
            assert (
                synthesizer_config.sampling_rate == SamplingRate.RATE_8000
            ), "MuLaw encoding only supports 8kHz sampling rate"
        self.filler_audios: List[FillerAudio] = []
        self.async_requestor = AsyncRequestor()
        self.total_chars: int = 0
        self.cost_per_char: Optional[float] = None

    @classmethod
    def get_voice_identifier(cls, synthesizer_config: SynthesizerConfigType) -> str:
        raise NotImplementedError

    async def empty_generator(self):
        yield SynthesisResult.ChunkResult(b"", True)

    def get_synthesizer_config(self) -> SynthesizerConfig:
        return self.synthesizer_config

    def get_typing_noise_filler_audio(self) -> FillerAudio:
        return FillerAudio(
            message=BaseMessage(text="<typing noise>"),
            audio_data=convert_wav(
                TYPING_NOISE_PATH,
                output_sample_rate=self.synthesizer_config.sampling_rate,
                output_encoding=self.synthesizer_config.audio_encoding,
            ),
            synthesizer_config=self.synthesizer_config,
            is_interruptible=True,
            seconds_per_chunk=2,
        )

    def get_cost(self) -> float:
        raise NotImplementedError

    async def set_filler_audios(self, filler_audio_config: FillerAudioConfig):
        if filler_audio_config.use_phrases:
            self.filler_audios = await self.get_phrase_filler_audios()
        elif filler_audio_config.use_typing_noise:
            self.filler_audios = [self.get_typing_noise_filler_audio()]

    async def get_phrase_filler_audios(self) -> List[FillerAudio]:
        return []

    def ready_synthesizer(self, chunk_size: int):
        pass

    # given the number of seconds the message was allowed to go until, where did we get in the message?
    @staticmethod
    def get_message_cutoff_from_total_response_length(
        synthesizer_config: SynthesizerConfig,
        message: BaseMessage,
        seconds: float,
        size_of_output: int,
    ) -> str:
        estimated_output_seconds = size_of_output / synthesizer_config.sampling_rate
        if not message.text:
            return message.text

        estimated_output_seconds_per_char = estimated_output_seconds / len(message.text)
        return message.text[: int(seconds / estimated_output_seconds_per_char)]

    @staticmethod
    def get_message_cutoff_from_voice_speed(
        message: BaseMessage, seconds: float, words_per_minute: int
    ) -> str:
        words_per_second = words_per_minute / 60
        estimated_words_spoken = math.floor(words_per_second * seconds)
        tokens = word_tokenize(message.text)
        return TreebankWordDetokenizer().detokenize(tokens[:estimated_words_spoken])

    async def get_cached_audio(
        self,
        message: BaseMessage,
    ) -> Optional[CachedAudio]:
        audio_cache = await AudioCache.safe_create()
        cache_phrase = message.cache_phrase or message.text.strip()
        audio_data = await audio_cache.get_audio(
            self.get_voice_identifier(self.synthesizer_config), cache_phrase
        )
        if audio_data is None:
            return None
        logger.info(f"Got cached audio for {cache_phrase}")

        trailing_silence_seconds = 0.0
        if isinstance(message, BotBackchannel):
            trailing_silence_seconds = message.trailing_silence_seconds
        return CachedAudio(message, audio_data, self.synthesizer_config, trailing_silence_seconds)

    async def create_speech_uncached(
        self,
        message: BaseMessage,
        chunk_size: int,
        is_first_text_chunk: bool = False,
        is_sole_text_chunk: bool = False,
    ) -> SynthesisResult:
        raise NotImplementedError

    async def create_speech(
        self,
        message: BaseMessage,
        chunk_size: int,
        is_first_text_chunk: bool = False,
        is_sole_text_chunk: bool = False,
    ) -> SynthesisResult:
        if isinstance(message, SilenceMessage):
            return SilenceAudio(
                message,
                self.synthesizer_config,
            ).create_synthesis_result(chunk_size)

        maybe_cached_audio = await self.get_cached_audio(message)
        if maybe_cached_audio is not None:
            return maybe_cached_audio.create_synthesis_result(chunk_size)
        return await self.create_speech_uncached(
            message,
            chunk_size,
            is_first_text_chunk=is_first_text_chunk,
            is_sole_text_chunk=is_sole_text_chunk,
        )

    async def chunk_result_generator_from_queue(self, chunk_queue: asyncio.Queue[Optional[bytes]]):
        while True:
            try:
                chunk = await chunk_queue.get()
                if chunk is None:
                    break
                yield SynthesisResult.ChunkResult(
                    chunk=chunk,
                    is_last_chunk=False,
                )
            except asyncio.CancelledError:
                break

    # @param file - a file-like object in wav format
    @staticmethod
    def create_synthesis_result_from_wav(
        synthesizer_config: SynthesizerConfig,
        file: Any,
        message: BaseMessage,
        chunk_size: int,
    ) -> SynthesisResult:
        output_bytes = convert_wav(
            file,
            output_sample_rate=synthesizer_config.sampling_rate,
            output_encoding=synthesizer_config.audio_encoding,
        )

        if synthesizer_config.should_encode_as_wav:
            chunk_transform = lambda chunk: encode_as_wav(chunk, synthesizer_config)  # noqa: E731

        else:
            chunk_transform = lambda chunk: chunk  # noqa: E731

        async def chunk_generator(output_bytes):
            for i in range(0, len(output_bytes), chunk_size):
                if i + chunk_size > len(output_bytes):
                    yield SynthesisResult.ChunkResult(chunk_transform(output_bytes[i:]), True)
                else:
                    yield SynthesisResult.ChunkResult(
                        chunk_transform(output_bytes[i : i + chunk_size]), False
                    )

        return SynthesisResult(
            chunk_generator(output_bytes),
            lambda seconds: BaseSynthesizer.get_message_cutoff_from_total_response_length(
                synthesizer_config, message, seconds, len(output_bytes)
            ),
        )

    async def experimental_mp3_streaming_output_generator(
        self,
        response: aiohttp.ClientResponse,
        chunk_size: int,
    ) -> AsyncGenerator[SynthesisResult.ChunkResult, None]:
        miniaudio_worker_input_queue: asyncio.Queue[Union[bytes, None]] = asyncio.Queue()
        miniaudio_worker_output_queue: asyncio.Queue[Tuple[bytes, bool]] = asyncio.Queue()
        miniaudio_worker = MiniaudioWorker(
            self.synthesizer_config,
            chunk_size,
            miniaudio_worker_input_queue,
            miniaudio_worker_output_queue,
        )
        miniaudio_worker.start()
        stream_reader = response.content

        # Create a task to send the mp3 chunks to the MiniaudioWorker's input queue in a separate loop
        async def send_chunks():
            async for chunk in stream_reader.iter_any():
                miniaudio_worker.consume_nonblocking(chunk)
            miniaudio_worker.consume_nonblocking(None)  # sentinel

        try:
            asyncio_create_task_with_done_error_log(send_chunks(), reraise_cancelled=True)

            # Await the output queue of the MiniaudioWorker and yield the wav chunks in another loop
            while True:
                # Get the wav chunk and the flag from the output queue of the MiniaudioWorker
                wav_chunk, is_last = await miniaudio_worker.output_queue.get()
                if self.synthesizer_config.should_encode_as_wav:
                    wav_chunk = encode_as_wav(wav_chunk, self.synthesizer_config)

                if self.synthesizer_config.audio_encoding == AudioEncoding.MULAW:
                    wav_chunk = audioop.lin2ulaw(wav_chunk, 2)

                yield SynthesisResult.ChunkResult(wav_chunk, is_last)
                # If this is the last chunk, break the loop
                if is_last:
                    break
        except asyncio.CancelledError:
            pass
        finally:
            miniaudio_worker.terminate()

    def _resample_chunk(
        self,
        chunk: bytes,
        current_sample_rate: int,
        target_sample_rate: int,
    ) -> bytes:
        resampled_chunk, _ = audioop.ratecv(
            chunk,
            2,
            1,
            current_sample_rate,
            target_sample_rate,
            None,
        )

        return resampled_chunk

    async def tear_down(self):
        pass



from loguru import logger

from vocode.streaming.models.transcriber import Transcription

LEARNING_RATE = 0.1
LENGTH_THRESHOLD = 5
SMOOTHING_FACTOR = 3
BASE_WPM = 150.0


class SpeedManager:
    """
    Uses the WPM of incoming transcriptions to update a speed coefficient to inform conversation orchestration.

    speed_coefficient is used in two places (currently):
    - DeepgramTranscriber endpointing - when we receive fast utterances from the user, we decrease the endpointing threshold to match the speed of the user
    - Interruption logic - if the user is speaking slowly, then when the agent is responding to an interrupt, we wait longer

    Initializes with a speed coefficient of 1.0, which corresponds to a WPM of 150.0 and uses a moving average to update the speed coefficient based on incoming transcriptions.
    """

    def __init__(self, speed_coefficient: float = 1.0):
        self.wpm_0 = BASE_WPM * speed_coefficient
        self.wpm = self.wpm_0
        self.speed_coefficient = speed_coefficient

    def update(self, transcription: Transcription):
        transcription_wpm = transcription.wpm()
        if transcription_wpm is not None:
            length = len(transcription.message.strip().split())
            p_t = min(
                1,
                LEARNING_RATE
                * ((length + SMOOTHING_FACTOR) / (LENGTH_THRESHOLD + SMOOTHING_FACTOR)),
            )
            self.wpm = self.wpm * (1 - p_t) + transcription_wpm * p_t
            self.speed_coefficient = self.wpm / BASE_WPM
            logger.info(f"Set speed coefficient to {self.speed_coefficient}")

    def get_speed_coefficient(self):
        return self.speed_coefficient

    def get_wpm(self):
        return self.wpm



"""Utilities for chat_gpt_agent, mostly around token counting for cost estimation purposes."""

import json
import textwrap
from typing import Any, Dict, List, NamedTuple, Optional

import tiktoken
from loguru import logger

# THE FOLLOWING CODE, UNTIL THE END MARKER, WERE RETRIEVED ON 9/13/2023 FROM
# THE OPENAI COOKBOOK UNDER THE MIT LICENSE.
# MIT License

# Copyright (c) 2023 OpenAI

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.


# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


# Used to count the amount of tokens Actions add to the billable cost
_FUNCTION_OVERHEAD_STR = """# Tools

## functions

namespace functions {

} // namespace functions"""

CHAT_GPT_MAX_TOKENS = {
    "gpt-3.5-turbo-0613": 4050,
    "gpt-3.5-turbo-16k-0613": 16340,
    "gpt-3.5-turbo-16k": 16340,
    "gpt-3.5-turbo": 16340,
    "gpt-3.5-turbo-1106": 16340,
    "gpt-3.5-turbo-0125": 16340,
    "gpt-4-0314": 8150,
    "gpt-4-32k-0314": 32700,
    "gpt-4-0613": 8150,
    "gpt-4-32k-0613": 32700,
    "gpt-4-0125-preview": 127940,
    "gpt-4-turbo": 127940,
    "gpt-4o": 127940,
    "gpt-4o-2024-05-13": 127940,
}


def get_chat_gpt_max_tokens(model_name: str):
    if model_name.startswith("ft:"):
        model_name = model_name.split(":")[1]

    if model_name in CHAT_GPT_MAX_TOKENS:
        return CHAT_GPT_MAX_TOKENS[model_name]

    return 4050


TokenizerInfo = NamedTuple(
    "TokenizerInfo",
    [
        ("encoding", tiktoken.Encoding),
        ("tokens_per_message", int),
        ("tokens_per_name", int),
    ],
)


def get_tokenizer_info(model: str) -> Optional[TokenizerInfo]:
    if "gpt-35-turbo" in model:
        model = "gpt-3.5-turbo"
    elif "gpt-4o" == model:
        model = "gpt-4o"
    elif "gpt4" in model or "gpt-4" in model:
        model = "gpt-4"
    try:
        encoding = tiktoken.encoding_for_model(model)
    except KeyError:
        logger.warning("Warning: model not found. Using cl100k_base encoding.")
        encoding = tiktoken.get_encoding("cl100k_base")
    if model in {
        "gpt-3.5-turbo-0613",
        "gpt-3.5-turbo-16k-0613",
        "gpt-4-0314",
        "gpt-4-32k-0314",
        "gpt-4-0613",
        "gpt-4-32k-0613",
    }:
        tokens_per_message = 3
        tokens_per_name = 1
    elif model == "gpt-3.5-turbo-0301":
        tokens_per_message = 4  # every message follows <|start|>{role/name}\n{content}<|end|>\n
        tokens_per_name = -1  # if there's a name, the role is omitted
    elif "gpt-3.5-turbo" in model:
        logger.debug(
            "Warning: gpt-3.5-turbo may update over time. Returning num tokens assuming gpt-3.5-turbo-0613."
        )
        tokens_per_message = 3
        tokens_per_name = 1
    elif "gpt-4" in model:
        logger.debug(
            "Warning: gpt-4 may update over time. Returning num tokens assuming gpt-4-0613."
        )
        tokens_per_message = 3
        tokens_per_name = 1
    else:
        return None

    return TokenizerInfo(
        encoding=encoding,
        tokens_per_message=tokens_per_message,
        tokens_per_name=tokens_per_name,
    )


def num_tokens_from_messages(messages: List[dict], model: str = "gpt-3.5-turbo-0613"):
    """Return the number of tokens used by a list of messages."""
    tokenizer_info = get_tokenizer_info(model)
    if tokenizer_info is None:
        raise NotImplementedError(
            f"""num_tokens_from_messages() is not implemented for model {model}. See https://github.com/openai/openai-python/blob/main/chatml.md for information on how messages are converted to tokens."""
        )
    num_tokens = 0
    for message in messages:
        num_tokens += tokenizer_info.tokens_per_message
        num_tokens += tokens_from_dict(
            encoding=tokenizer_info.encoding,
            d=message,
            tokens_per_name=tokenizer_info.tokens_per_name,
        )
    num_tokens += 3  # every reply is primed with <|start|>assistant<|message|>
    return num_tokens


# END OF OPENAI COOKBOOK CODE AND GIVEN MIT LICENSE.


def tokens_from_dict(encoding: tiktoken.Encoding, d: Dict[str, Any], tokens_per_name: int) -> int:
    """Return the number of OpenAI tokens in a dict."""
    num_tokens: int = 0
    for key, value in d.items():
        if value is None:
            continue
        if isinstance(value, str):
            num_tokens += len(encoding.encode(value))
            if key == "name":
                num_tokens += tokens_per_name
        elif isinstance(value, dict):
            num_tokens += tokens_from_dict(
                encoding=encoding, d=value, tokens_per_name=tokens_per_name
            )

    return num_tokens


def num_tokens_from_functions(functions: List[dict] | None, model="gpt-3.5-turbo-0613") -> int:
    """Return the number of tokens used by a list of functions."""
    if not functions:
        return 0

    try:
        encoding = tiktoken.encoding_for_model(model)
    except KeyError:
        logger.warning("Warning: model not found. Using cl100k_base encoding.")
        encoding = tiktoken.get_encoding("cl100k_base")

    function_overhead = 3 + len(encoding.encode(_FUNCTION_OVERHEAD_STR))

    return function_overhead + sum(
        len(encoding.encode(_format_func_into_prompt_str(func=f))) for f in functions
    )


# Calculates the amount of tokens added to a given OpenAI prompt for functions
# specifically for billing purposes
def _format_func_into_prompt_str(func) -> str:
    def resolve_ref(schema):
        if schema.get("$ref") is not None:
            ref = schema["$ref"][14:]
            schema = json_schema["definitions"][ref]
        return schema

    def format_schema(schema, indent):
        schema = resolve_ref(schema)
        if "enum" in schema:
            return format_enum(schema, indent)
        elif schema["type"] == "object":
            return format_object(schema, indent)
        elif schema["type"] == "integer":
            return "number"
        elif schema["type"] == "boolean":
            return "boolean"
        elif schema["type"] in ["string", "number"]:
            return schema["type"]
        elif schema["type"] == "array":
            return format_schema(schema["items"], indent) + "[]"
        else:
            raise ValueError("unknown schema type " + schema["type"])

    def format_enum(schema, indent):
        return " | ".join(json.dumps(o) for o in schema["enum"])

    def format_object(schema, indent):
        result = "{\n"
        if "properties" not in schema or len(schema["properties"]) == 0:
            if schema.get("additionalProperties", False):
                return "object"
            return None
        for key, value in schema["properties"].items():
            value = resolve_ref(value)
            value_rendered = format_schema(value, indent + 1)
            if value_rendered is None:
                continue
            if "description" in value and indent == 0:
                for line in textwrap.dedent(value["description"]).strip().split("\n"):
                    result += f"{'  '*indent}// {line}\n"
            optional = "" if key in schema.get("required", {}) else "?"
            comment = (
                "" if value.get("default") is None else f" // default: {format_default(value)}"
            )
            result += f"{'  '*indent}{key}{optional}: {value_rendered},{comment}\n"
        result += ("  " * (indent - 1)) + "}"
        return result

    def format_default(schema):
        v = schema["default"]
        if schema["type"] == "number":
            return f"{v:.1f}" if float(v).is_integer() else str(v)
        else:
            return str(v)

    json_schema = func["parameters"]
    result = f"// {func['description']}\ntype {func['name']} = ("
    formatted = format_object(json_schema, 0)
    if formatted is not None:
        result += "_: " + formatted
    result += ") => any;\n\n"
    return result



from typing import Type

from loguru import logger
from pydantic.v1 import BaseModel, Field

from vocode.streaming.action.phone_call_action import (
    TwilioPhoneConversationAction,
    VonagePhoneConversationAction,
)
from vocode.streaming.models.actions import ActionConfig as VocodeActionConfig
from vocode.streaming.models.actions import ActionInput, ActionOutput
from vocode.streaming.utils.state_manager import (
    TwilioPhoneConversationStateManager,
    VonagePhoneConversationStateManager,
)


class DTMFParameters(BaseModel):
    buttons: str = Field(..., description="The buttons to press.")


class DTMFResponse(BaseModel):
    success: bool


class DTMFVocodeActionConfig(VocodeActionConfig, type="action_dtmf"):  # type: ignore
    def action_attempt_to_string(self, input: ActionInput) -> str:
        assert isinstance(input.params, DTMFParameters)
        return "Attempting to press numbers: " f"{list(input.params.buttons)}"

    def action_result_to_string(self, input: ActionInput, output: ActionOutput) -> str:
        assert isinstance(input.params, DTMFParameters)
        assert isinstance(output.response, DTMFResponse)
        return f"Pressed numbers {list(input.params.buttons)} successfully"


FUNCTION_DESCRIPTION = "Presses a string numbers using DTMF tones."


class VonageDTMF(
    VonagePhoneConversationAction[DTMFVocodeActionConfig, DTMFParameters, DTMFResponse]
):
    description: str = FUNCTION_DESCRIPTION
    parameters_type: Type[DTMFParameters] = DTMFParameters
    response_type: Type[DTMFResponse] = DTMFResponse
    conversation_state_manager: VonagePhoneConversationStateManager

    def __init__(self, action_config: DTMFVocodeActionConfig):
        super().__init__(action_config, quiet=True)

    async def run(self, action_input: ActionInput[DTMFParameters]) -> ActionOutput[DTMFResponse]:
        buttons = action_input.params.buttons
        vonage_client = self.conversation_state_manager.create_vonage_client()
        await vonage_client.send_dtmf(
            vonage_uuid=self.get_vonage_uuid(action_input), digits=buttons
        )

        return ActionOutput(
            action_type=action_input.action_config.type,
            response=DTMFResponse(success=True),
        )


class TwilioDTMF(
    TwilioPhoneConversationAction[DTMFVocodeActionConfig, DTMFParameters, DTMFResponse]
):
    description: str = FUNCTION_DESCRIPTION
    parameters_type: Type[DTMFParameters] = DTMFParameters
    response_type: Type[DTMFResponse] = DTMFResponse
    conversation_state_manager: TwilioPhoneConversationStateManager

    def __init__(self, action_config: DTMFVocodeActionConfig):
        super().__init__(
            action_config,
            quiet=True,
        )

    async def run(self, action_input: ActionInput[DTMFParameters]) -> ActionOutput[DTMFResponse]:
        logger.error("DTMF not yet supported with Twilio")
        return ActionOutput(
            action_type=action_input.action_config.type,
            response=DTMFResponse(success=False),
        )



from enum import Enum
from typing import Optional

from vocode.streaming.models.model import TypedModel


class Sender(str, Enum):
    HUMAN = "human"
    BOT = "bot"
    ACTION_WORKER = "action_worker"
    VECTOR_DB = "vector_db"
    CONFERENCE = "conference"


class EventType(str, Enum):
    TRANSCRIPT = "event_transcript"
    TRANSCRIPT_COMPLETE = "event_transcript_complete"
    PHONE_CALL_CONNECTED = "event_phone_call_connected"
    PHONE_CALL_ENDED = "event_phone_call_ended"
    PHONE_CALL_DID_NOT_CONNECT = "event_phone_call_did_not_connect"
    RECORDING = "event_recording"
    ACTION = "event_action"


class Event(TypedModel):
    conversation_id: str


class PhoneCallConnectedEvent(Event, type=EventType.PHONE_CALL_CONNECTED):  # type: ignore
    to_phone_number: str
    from_phone_number: str


class PhoneCallEndedEvent(Event, type=EventType.PHONE_CALL_ENDED):  # type: ignore
    conversation_minutes: float = 0


class PhoneCallDidNotConnectEvent(Event, type=EventType.PHONE_CALL_DID_NOT_CONNECT):  # type: ignore
    telephony_status: str


class RecordingEvent(Event, type=EventType.RECORDING):  # type: ignore
    recording_url: str


class ActionEvent(Event, type=EventType.ACTION):  # type: ignore
    action_input: Optional[dict] = None
    action_output: Optional[dict] = None



import asyncio
import audioop
import os
from typing import AsyncGenerator, AsyncIterator, Optional

import numpy as np
from loguru import logger
from pyht import AsyncClient
from pyht.client import CongestionCtrl, TTSOptions
from pyht.protos import api_pb2

from vocode.streaming.models.audio import AudioEncoding
from vocode.streaming.models.message import BaseMessage
from vocode.streaming.models.synthesizer import PlayHtSynthesizerConfig
from vocode.streaming.synthesizer.base_synthesizer import SynthesisResult
from vocode.streaming.synthesizer.play_ht_synthesizer import (
    PlayHtSynthesizer as VocodePlayHtSynthesizer,
)
from vocode.streaming.synthesizer.synthesizer_utils import split_text
from vocode.streaming.utils import generate_from_async_iter_with_lookahead, generate_with_is_last
from vocode.streaming.utils.create_task import asyncio_create_task_with_done_error_log

PLAY_HT_ON_PREM_ADDR = os.environ.get("VOCODE_PLAYHT_ON_PREM_ADDR", None)
PLAY_HT_V2_MAX_CHARS = 200
EXPERIMENTAL_VOICE_AMPLITUDE_THRESHOLD = 200


class PlayHtSynthesizerV2(VocodePlayHtSynthesizer):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.playht_client_saas = AsyncClient(
            user_id=self.user_id,
            api_key=self.api_key,
        )
        self.playht_client_on_prem = None

        if self.synthesizer_config.on_prem:
            logger.info(f"Creating on-prem PlayHT with gRPC address {PLAY_HT_ON_PREM_ADDR}")
            advanced_options = AsyncClient.AdvancedOptions(
                grpc_addr=PLAY_HT_ON_PREM_ADDR,
                fallback_enabled=True,
                congestion_ctrl=CongestionCtrl.STATIC_MAR_2023,
            )

            self.playht_client_on_prem = AsyncClient(
                user_id=self.user_id,
                api_key=self.api_key,
                advanced=advanced_options,
            )

        if self.synthesizer_config.audio_encoding == AudioEncoding.MULAW:
            audio_format = api_pb2.FORMAT_MULAW
        elif self.synthesizer_config.audio_encoding == AudioEncoding.LINEAR16:
            audio_format = api_pb2.FORMAT_WAV

        sample_rate = self.synthesizer_config.sampling_rate
        self.playht_options = TTSOptions(
            voice=self.synthesizer_config.voice_id,
            # PlayHT runs significantly slower when sampling rate is not 24KHz
            # 24KHz is the default for PlayHT
            sample_rate=sample_rate if sample_rate > 24000 else 24000,
            speed=self.synthesizer_config.speed if self.synthesizer_config.speed else 1,
            format=audio_format,
            text_guidance=self.synthesizer_config.text_guidance,
            voice_guidance=self.synthesizer_config.voice_guidance,
            temperature=self.synthesizer_config.temperature,
            top_p=self.synthesizer_config.top_p,
        )
        if self.synthesizer_config.quality:
            self.playht_options.quality = self.synthesizer_config.quality

    @property
    def playht_client(self) -> AsyncClient:
        if self.playht_client_on_prem is not None:
            return self.playht_client_on_prem
        return self.playht_client_saas

    async def create_speech_uncached(
        self,
        message: BaseMessage,
        chunk_size: int,
        is_first_text_chunk: bool = False,
        is_sole_text_chunk: bool = False,
    ) -> SynthesisResult:

        self.total_chars += len(message.text)
        chunk_queue: asyncio.Queue[Optional[bytes]] = asyncio.Queue()
        asyncio_create_task_with_done_error_log(
            self.get_chunks(
                message,
                chunk_size,
                chunk_queue,
                cut_leading_silence=not is_first_text_chunk
                and self.synthesizer_config.experimental_remove_silence,
                cut_trailing_silence=not is_sole_text_chunk
                and self.synthesizer_config.experimental_remove_silence,
            ),
        )

        return SynthesisResult(
            self.chunk_result_generator_from_queue(chunk_queue),
            lambda seconds: self.get_message_cutoff_from_voice_speed(
                message,
                seconds,
                self.words_per_minute,
            ),
        )

    def _contains_voice_experimental(self, chunk: bytes):
        pcm = np.frombuffer(
            (
                audioop.ulaw2lin(chunk, 2)
                if self.synthesizer_config.audio_encoding == AudioEncoding.MULAW
                else chunk
            ),
            dtype=np.int16,
        )
        return np.max(np.abs(pcm)) > EXPERIMENTAL_VOICE_AMPLITUDE_THRESHOLD

    @staticmethod
    def _enumerate_by_chunk_size(buffer: bytes, chunk_size: int):
        for buffer_idx in range(0, len(buffer) - chunk_size, chunk_size):
            yield buffer_idx, buffer[buffer_idx : buffer_idx + chunk_size]

    async def _downsample_from_24khz(self, chunk: bytes) -> bytes:
        if self.synthesizer_config.audio_encoding == AudioEncoding.MULAW:
            return await self._downsample_mulaw(chunk)
        elif self.synthesizer_config.audio_encoding == AudioEncoding.LINEAR16:
            return await self._downsample_pcm(chunk)
        else:
            raise Exception(f"Unsupported audio format: {self.synthesizer_config.audio_encoding}")

    async def _downsample_pcm(self, chunk: bytes) -> bytes:
        downsampled_chunk, _ = audioop.ratecv(
            chunk,
            2,
            1,
            24000,
            self.synthesizer_config.sampling_rate,
            None,
        )

        return downsampled_chunk

    async def _downsample_mulaw(self, chunk: bytes) -> bytes:
        pcm_data = audioop.ulaw2lin(chunk, 2)
        downsampled_pcm_data = await self._downsample_pcm(pcm_data)
        downsampled_chunk = audioop.lin2ulaw(downsampled_pcm_data, 2)
        return downsampled_chunk

    async def downsample_async_generator(self, async_gen: AsyncGenerator[bytes, None]):
        async for play_ht_chunk in async_gen:
            if self.synthesizer_config.sampling_rate >= 24000:
                yield play_ht_chunk
            else:
                downsampled_chunk = await self._downsample_from_24khz(play_ht_chunk)
                yield downsampled_chunk

    async def _cut_leading_trailing_silence(
        self,
        async_iter: AsyncIterator[bytes],
        chunk_size: int,
        cut_leading_silence: bool = True,
        cut_trailing_silence: bool = True,
    ) -> AsyncGenerator[bytes, None]:
        buffer: bytearray = bytearray()

        async def generate_chunks(
            play_ht_chunk: bytes,
            cut_leading_silence=False,
        ) -> AsyncGenerator[bytes, None]:
            """Yields chunks of size chunk_size from play_ht_chunk and leaves the remainder in buffer.

            If cut_leading_silence is True, does not yield chunks until it detects voice.
            """
            nonlocal buffer

            buffer.extend(play_ht_chunk)
            detected_voice = False
            for buffer_idx, chunk in self._enumerate_by_chunk_size(buffer, chunk_size):
                if cut_leading_silence and not detected_voice:
                    if self._contains_voice_experimental(chunk):
                        detected_voice = True
                        yield chunk
                    if detected_voice:
                        logger.debug(f"Cut off {buffer_idx} bytes of leading silence")
                else:
                    yield chunk
            buffer = buffer[len(buffer) - (len(buffer) % chunk_size) :]

        async def _cut_out_trailing_silence(
            trailing_chunk: bytes,
        ) -> AsyncGenerator[bytes, None]:
            """Yields chunks of size chunk_size from trailing_chunk until it detects silence."""
            for buffer_idx, chunk in self._enumerate_by_chunk_size(trailing_chunk, chunk_size):
                if not self._contains_voice_experimental(chunk):
                    logger.debug(
                        f"Cutting off {len(trailing_chunk) - buffer_idx} bytes of trailing silence",
                    )
                    break
                yield chunk

        # Yield from the first audio chunk, no matter what, for latency
        try:
            play_ht_chunk = await async_iter.__anext__()
        except StopAsyncIteration:
            return

        async for chunk in generate_chunks(play_ht_chunk, cut_leading_silence=cut_leading_silence):
            yield chunk

        async for lookahead_buffer, is_last in generate_with_is_last(
            generate_from_async_iter_with_lookahead(async_iter, 2),
        ):
            if not is_last:
                async for chunk in generate_chunks(lookahead_buffer[0]):
                    yield chunk
            else:
                trailing_chunk = b"".join(lookahead_buffer)
                async for chunk in (
                    _cut_out_trailing_silence(trailing_chunk)
                    if cut_trailing_silence
                    else generate_chunks(trailing_chunk)
                ):
                    yield chunk

    async def get_chunks(
        self,
        message: BaseMessage,
        chunk_size: int,
        chunk_queue: asyncio.Queue[Optional[bytes]],
        cut_leading_silence: bool,
        cut_trailing_silence: bool,
    ):
        buffer = bytearray()
        try:
            playht_bytes_generators = [
                self.playht_client.tts(
                    text,
                    self.playht_options,
                )
                for text in split_text(
                    string_to_split=message.text,
                    max_text_length=PLAY_HT_V2_MAX_CHARS,
                )
            ]
            downsampled_generators = [
                self.downsample_async_generator(gen) for gen in playht_bytes_generators
            ]

            for async_gen in downsampled_generators:
                async_iter = async_gen.__aiter__()
                if (
                    self.synthesizer_config.audio_encoding == AudioEncoding.LINEAR16
                ):  # skip the first chunk, which contains wav header
                    await async_iter.__anext__()
                if not cut_trailing_silence and not cut_leading_silence:
                    while True:
                        try:
                            play_ht_chunk = await async_iter.__anext__()
                        except StopAsyncIteration:
                            break

                        buffer.extend(play_ht_chunk)
                        for _, chunk in self._enumerate_by_chunk_size(buffer, chunk_size):
                            chunk_queue.put_nowait(chunk)
                        buffer = buffer[len(buffer) - (len(buffer) % chunk_size) :]
                    if len(buffer) > 0:
                        chunk_queue.put_nowait(buffer)
                else:
                    async for chunk in self._cut_leading_trailing_silence(
                        async_iter,
                        chunk_size,
                        cut_leading_silence=cut_leading_silence,
                        cut_trailing_silence=cut_trailing_silence,
                    ):
                        chunk_queue.put_nowait(chunk)
        except asyncio.CancelledError:
            pass
        finally:
            chunk_queue.put_nowait(None)  # treated as sentinel

    @classmethod
    def get_voice_identifier(cls, synthesizer_config: PlayHtSynthesizerConfig):
        return ":".join(
            (
                "play_ht_v2",
                synthesizer_config.voice_id,
                str(synthesizer_config.user_id),
                str(synthesizer_config.speed),
                str(synthesizer_config.seed),
                str(synthesizer_config.temperature),
                synthesizer_config.audio_encoding,
            ),
        )

    async def tear_down(self):
        await self.playht_client.close()
        await super().tear_down()



from __future__ import annotations

import asyncio
import queue
import random
import re
import threading
import time
import typing
from typing import (
    Any,
    AsyncGenerator,
    Awaitable,
    Callable,
    Generic,
    Iterator,
    List,
    Optional,
    Tuple,
    TypeVar,
    Union,
)

import sentry_sdk
from loguru import logger
from sentry_sdk.tracing import Span

from vocode.streaming.action.worker import ActionsWorker
from vocode.streaming.agent.base_agent import (
    AgentInput,
    AgentResponse,
    AgentResponseFillerAudio,
    AgentResponseMessage,
    AgentResponseStop,
    BaseAgent,
    TranscriptionAgentInput,
)
from vocode.streaming.agent.chat_gpt_agent import ChatGPTAgent
from vocode.streaming.constants import (
    ALLOWED_IDLE_TIME,
    CHECK_HUMAN_PRESENT_MESSAGE_CHOICES,
    PER_CHUNK_ALLOWANCE_SECONDS,
    TEXT_TO_SPEECH_CHUNK_SIZE_SECONDS,
)
from vocode.streaming.models.actions import EndOfTurn
from vocode.streaming.models.agent import FillerAudioConfig
from vocode.streaming.models.events import Sender
from vocode.streaming.models.message import BaseMessage, BotBackchannel, LLMToken, SilenceMessage
from vocode.streaming.models.transcriber import TranscriberConfig, Transcription
from vocode.streaming.models.transcript import Message, Transcript, TranscriptCompleteEvent
from vocode.streaming.output_device.base_output_device import BaseOutputDevice
from vocode.streaming.synthesizer.base_synthesizer import (
    BaseSynthesizer,
    FillerAudio,
    SynthesisResult,
)
from vocode.streaming.synthesizer.input_streaming_synthesizer import InputStreamingSynthesizer
from vocode.streaming.transcriber.base_transcriber import BaseTranscriber
from vocode.streaming.transcriber.deepgram_transcriber import DeepgramTranscriber
from vocode.streaming.utils import create_conversation_id, get_chunk_size_per_second
from vocode.streaming.utils.create_task import asyncio_create_task_with_done_error_log
from vocode.streaming.utils.events_manager import EventsManager
from vocode.streaming.utils.speed_manager import SpeedManager
from vocode.streaming.utils.state_manager import ConversationStateManager
from vocode.streaming.utils.worker import (
    AsyncQueueWorker,
    InterruptibleAgentResponseEvent,
    InterruptibleAgentResponseWorker,
    InterruptibleEvent,
    InterruptibleEventFactory,
)
from vocode.utils.sentry_utils import (
    CustomSentrySpans,
    complete_span_by_op,
    sentry_create_span,
    synthesizer_base_name_if_should_report_to_sentry,
)

BACKCHANNEL_PATTERNS = [
    r"m+-?hm+",
    r"m+",
    r"oh+",
    r"ah+",
    r"um+",
    r"uh+",
    "yes",
    "sure",
    "quite",
    "right",
    "really",
    "good heavens",
    "i see",
    "of course",
    "oh dear",
    "oh god",
    "thats nice",
    "thats not bad",
    "thats right",
    r"yeah+",
    "makes sense",
]
LOW_INTERRUPT_SENSITIVITY_BACKCHANNEL_UTTERANCE_LENGTH_THRESHOLD = 3


OutputDeviceType = TypeVar("OutputDeviceType", bound=BaseOutputDevice)


class StreamingConversation(Generic[OutputDeviceType]):
    class QueueingInterruptibleEventFactory(InterruptibleEventFactory):
        def __init__(self, conversation: "StreamingConversation"):
            self.conversation = conversation

        def create_interruptible_event(
            self,
            payload: Any,
            is_interruptible: bool = True,
        ) -> InterruptibleEvent[Any]:
            interruptible_event: InterruptibleEvent = super().create_interruptible_event(
                payload,
                is_interruptible,
            )
            self.conversation.interruptible_events.put_nowait(interruptible_event)
            return interruptible_event

        def create_interruptible_agent_response_event(
            self,
            payload: Any,
            is_interruptible: bool = True,
            agent_response_tracker: Optional[asyncio.Event] = None,
        ) -> InterruptibleAgentResponseEvent:
            interruptible_event = super().create_interruptible_agent_response_event(
                payload,
                is_interruptible=is_interruptible,
                agent_response_tracker=agent_response_tracker,
            )
            self.conversation.interruptible_events.put_nowait(interruptible_event)
            return interruptible_event

    class TranscriptionsWorker(AsyncQueueWorker):
        """Processes all transcriptions: sends an interrupt if needed
        and sends final transcriptions to the output queue"""

        def __init__(
            self,
            input_queue: asyncio.Queue[Transcription],
            output_queue: asyncio.Queue[InterruptibleEvent[AgentInput]],
            conversation: "StreamingConversation",
            interruptible_event_factory: InterruptibleEventFactory,
        ):
            super().__init__(input_queue, output_queue)
            self.input_queue = input_queue
            self.output_queue = output_queue
            self.conversation = conversation
            self.interruptible_event_factory = interruptible_event_factory
            self.in_interrupt_endpointing_config = False
            self.deepgram_transcriber: Optional[DeepgramTranscriber] = None
            if isinstance(self.conversation.transcriber, DeepgramTranscriber):
                self.deepgram_transcriber = self.conversation.transcriber
            self.has_associated_ignored_utterance: bool = False
            self.has_associated_unignored_utterance: bool = False
            self.human_backchannels_buffer: List[Transcription] = []
            self.ignore_next_message: bool = False

        def should_ignore_utterance(self, transcription: Transcription):
            if self.has_associated_unignored_utterance:
                return False
            bot_still_speaking = self.is_bot_still_speaking()
            if self.has_associated_ignored_utterance or bot_still_speaking:
                logger.info(
                    f"Associated ignored utterance: {self.has_associated_ignored_utterance}. Bot still speaking: {bot_still_speaking}"
                )
                return self.is_transcription_backchannel(transcription)
            return False

        def is_transcription_backchannel(self, transcription: Transcription):
            num_words = len(transcription.message.strip().split())
            if (
                self.conversation.agent.get_agent_config().interrupt_sensitivity == "high"
                and num_words >= 1
            ):
                logger.info(f"High interrupt sensitivity; {num_words} word(s) not a backchannel")
                return False

            if num_words <= LOW_INTERRUPT_SENSITIVITY_BACKCHANNEL_UTTERANCE_LENGTH_THRESHOLD:
                return True
            cleaned = re.sub("[^\w\s]", "", transcription.message).strip().lower()
            return any(re.fullmatch(regex, cleaned) for regex in BACKCHANNEL_PATTERNS)

        def _most_recent_transcript_messages(self) -> Iterator[Message]:
            return (
                event_log
                for event_log in reversed(self.conversation.transcript.event_logs)
                if isinstance(event_log, Message)
            )

        def get_maybe_last_transcript_event_log(self) -> Optional[Message]:
            return next(self._most_recent_transcript_messages(), None)

        def is_bot_in_medias_res(self):
            last_message = self.get_maybe_last_transcript_event_log()
            return (
                last_message is not None
                and not last_message.is_backchannel
                and last_message.sender == Sender.BOT
                and not last_message.is_final
                and last_message.text.strip() != ""
            )

        def is_bot_still_speaking(self):  # in_medias_res OR bot has more utterances
            transcript_messages_iter = self._most_recent_transcript_messages()
            last_message, second_to_last_message = next(transcript_messages_iter, None), next(
                transcript_messages_iter, None
            )

            is_first_bot_message = (
                second_to_last_message is None or second_to_last_message.sender == Sender.HUMAN
            )

            return (
                last_message is not None
                and not last_message.is_backchannel
                and last_message.sender == Sender.BOT
                and (not last_message.is_final or not last_message.is_end_of_turn)
                and not (is_first_bot_message and last_message.text.strip() == "")
            )

        async def process(self, transcription: Transcription):
            self.conversation.mark_last_action_timestamp()
            if transcription.message.strip() == "":
                logger.info("Ignoring empty transcription")
                return
            # ignore utterances during the initial message but still add them to the transcript
            initial_message_ongoing = not self.conversation.initial_message_tracker.is_set()
            if initial_message_ongoing or self.should_ignore_utterance(transcription):
                logger.info(
                    f"Ignoring utterance: {transcription.message}. IMO: {initial_message_ongoing}"
                )
                self.has_associated_ignored_utterance = (
                    not transcription.is_final  # if it's final, we're done with this backchannel
                )
                if transcription.is_final:
                    # for all ignored backchannels, store them to be added to the transcript later
                    self.human_backchannels_buffer.append(transcription)
                return
            if self.ignore_next_message and transcription.is_final:
                # TODO: delete this once transcription reset is implemented for processing conference voicemail
                # Push human message to transcript but do not respond
                self.has_associated_ignored_utterance = False
                agent_response_tracker = None
                self.ignore_next_message = False
                return
            if transcription.is_final:
                if (
                    self.deepgram_transcriber is not None
                    and self.deepgram_transcriber.is_first_transcription
                ):
                    logger.debug(
                        "Switching to non-first transcription endpointing config if configured"
                    )
                    self.deepgram_transcriber.is_first_transcription = False
                logger.debug(
                    "Got transcription: {}, confidence: {}, wpm: {}".format(
                        transcription.message,
                        transcription.confidence,
                        transcription.wpm(),
                    )
                )
            if not self.conversation.is_human_speaking:
                self.conversation.current_transcription_is_interrupt = (
                    self.conversation.broadcast_interrupt()
                )
                self.has_associated_unignored_utterance = not transcription.is_final
                if self.conversation.current_transcription_is_interrupt:
                    logger.debug("sent interrupt")
                logger.debug("Human started speaking")

            transcription.is_interrupt = self.conversation.current_transcription_is_interrupt
            self.conversation.is_human_speaking = not transcription.is_final
            if transcription.is_final:
                self.has_associated_ignored_utterance = False
                self.has_associated_unignored_utterance = False
                agent_response_tracker = None

                # clear out backchannels and add to the transcript
                for human_backchannel in self.human_backchannels_buffer:
                    self.conversation.transcript.add_human_message(
                        text=human_backchannel.message,
                        conversation_id=self.conversation.id,
                        is_backchannel=True,
                    )
                self.human_backchannels_buffer = []

                if transcription.is_interrupt:
                    transcription.bot_was_in_medias_res = self.is_bot_in_medias_res()
                    logger.debug(
                        f"Bot is {'not ' if not transcription.bot_was_in_medias_res else ''}in medias res"
                    )

                self.conversation.speed_manager.update(transcription)

                self.conversation.warmup_synthesizer()

                # we use getattr here to avoid the dependency cycle between PhoneConversation and StreamingConversation
                event = self.interruptible_event_factory.create_interruptible_event(
                    TranscriptionAgentInput(
                        transcription=transcription,
                        conversation_id=self.conversation.id,
                        vonage_uuid=getattr(self.conversation, "vonage_uuid", None),
                        twilio_sid=getattr(self.conversation, "twilio_sid", None),
                        agent_response_tracker=agent_response_tracker,
                    ),
                )
                self.output_queue.put_nowait(event)

    class FillerAudioWorker(InterruptibleAgentResponseWorker):
        """
        - Waits for a configured number of seconds and then sends filler audio to the output
        - Exposes wait_for_filler_audio_to_finish() which the AgentResponsesWorker waits on before
          sending responses to the output queue
        """

        def __init__(
            self,
            input_queue: asyncio.Queue[InterruptibleAgentResponseEvent[FillerAudio]],
            conversation: "StreamingConversation",
        ):
            super().__init__(input_queue=input_queue)
            self.input_queue = input_queue
            self.conversation = conversation
            self.current_filler_seconds_per_chunk: Optional[int] = None
            self.filler_audio_started_event: Optional[threading.Event] = None

        async def wait_for_filler_audio_to_finish(self):
            if self.filler_audio_started_event is None or not self.filler_audio_started_event.set():
                logger.debug(
                    "Not waiting for filler audio to finish since we didn't send any chunks",
                )
                return
            if self.interruptible_event and isinstance(
                self.interruptible_event,
                InterruptibleAgentResponseEvent,
            ):
                await self.interruptible_event.agent_response_tracker.wait()

        def interrupt_current_filler_audio(self):
            return self.interruptible_event and self.interruptible_event.interrupt()

        async def process(self, item: InterruptibleAgentResponseEvent[FillerAudio]):
            try:
                filler_audio = item.payload
                assert self.conversation.filler_audio_config is not None
                filler_synthesis_result = filler_audio.create_synthesis_result()
                self.current_filler_seconds_per_chunk = filler_audio.seconds_per_chunk
                silence_threshold = self.conversation.filler_audio_config.silence_threshold_seconds
                await asyncio.sleep(silence_threshold)
                logger.debug("Sending filler audio to output")
                self.filler_audio_started_event = threading.Event()
                await self.conversation.send_speech_to_output(
                    filler_audio.message.text,
                    filler_synthesis_result,
                    item.interruption_event,
                    filler_audio.seconds_per_chunk,
                    started_event=self.filler_audio_started_event,
                )
                item.agent_response_tracker.set()
            except asyncio.CancelledError:
                pass

    class AgentResponsesWorker(InterruptibleAgentResponseWorker):
        """Runs Synthesizer.create_speech and sends the SynthesisResult to the output queue"""

        def __init__(
            self,
            input_queue: asyncio.Queue[InterruptibleAgentResponseEvent[AgentResponse]],
            output_queue: asyncio.Queue[
                InterruptibleAgentResponseEvent[
                    Tuple[Union[BaseMessage, EndOfTurn], Optional[SynthesisResult]]
                ]
            ],
            conversation: "StreamingConversation",
            interruptible_event_factory: InterruptibleEventFactory,
        ):
            super().__init__(
                input_queue=input_queue,
                output_queue=output_queue,
            )
            self.input_queue = input_queue
            self.output_queue = output_queue
            self.conversation = conversation
            self.interruptible_event_factory = interruptible_event_factory
            self.chunk_size = self.conversation._get_synthesizer_chunk_size()
            self.last_agent_response_tracker: Optional[asyncio.Event] = None
            self.is_first_text_chunk = True

        def send_filler_audio(self, agent_response_tracker: Optional[asyncio.Event]):
            assert self.conversation.filler_audio_worker is not None
            logger.debug("Sending filler audio")
            if self.conversation.synthesizer.filler_audios:
                filler_audio = random.choice(self.conversation.synthesizer.filler_audios)
                logger.debug(f"Chose {filler_audio.message.text}")
                event = self.interruptible_event_factory.create_interruptible_agent_response_event(
                    filler_audio,
                    is_interruptible=filler_audio.is_interruptible,
                    agent_response_tracker=agent_response_tracker,
                )
                self.conversation.filler_audio_worker.consume_nonblocking(event)
            else:
                logger.debug("No filler audio available for synthesizer")

        async def process(self, item: InterruptibleAgentResponseEvent[AgentResponse]):
            if not self.conversation.synthesis_enabled:
                logger.debug("Synthesis disabled, not synthesizing speech")
                return
            try:
                agent_response = item.payload
                if isinstance(agent_response, AgentResponseFillerAudio):
                    self.send_filler_audio(item.agent_response_tracker)
                    return
                if isinstance(agent_response, AgentResponseStop):
                    logger.debug("Agent requested to stop")
                    if self.last_agent_response_tracker is not None:
                        await self.last_agent_response_tracker.wait()
                    item.agent_response_tracker.set()
                    await self.conversation.mark_terminated(bot_disconnect=True)
                    return

                agent_response_message = typing.cast(AgentResponseMessage, agent_response)

                if self.conversation.filler_audio_worker is not None:
                    if self.conversation.filler_audio_worker.interrupt_current_filler_audio():
                        await self.conversation.filler_audio_worker.wait_for_filler_audio_to_finish()

                if isinstance(agent_response_message.message, EndOfTurn):
                    logger.debug("Sending end of turn")
                    if isinstance(self.conversation.synthesizer, InputStreamingSynthesizer):
                        await self.conversation.synthesizer.handle_end_of_turn()
                    self.produce_interruptible_agent_response_event_nonblocking(
                        (agent_response_message.message, None),
                        is_interruptible=item.is_interruptible,
                        agent_response_tracker=item.agent_response_tracker,
                    )
                    self.is_first_text_chunk = True
                    return

                synthesizer_base_name: Optional[str] = (
                    synthesizer_base_name_if_should_report_to_sentry(self.conversation.synthesizer)
                )
                create_speech_span: Optional[Span] = None
                ttft_span: Optional[Span] = None
                synthesis_span: Optional[Span] = None
                if synthesizer_base_name and agent_response_message.is_first:
                    complete_span_by_op(CustomSentrySpans.LANGUAGE_MODEL_TIME_TO_FIRST_TOKEN)

                    sentry_create_span(
                        sentry_callable=sentry_sdk.start_span,
                        op=CustomSentrySpans.SYNTHESIS_TIME_TO_FIRST_TOKEN,
                    )

                    synthesis_span = sentry_create_span(
                        sentry_callable=sentry_sdk.start_span,
                        op=f"{synthesizer_base_name}{CustomSentrySpans.SYNTHESIZER_SYNTHESIS_TOTAL}",
                    )
                    if synthesis_span:
                        ttft_span = sentry_create_span(
                            sentry_callable=synthesis_span.start_child,
                            op=f"{synthesizer_base_name}{CustomSentrySpans.SYNTHESIZER_TIME_TO_FIRST_TOKEN}",
                        )
                    if ttft_span:
                        create_speech_span = sentry_create_span(
                            sentry_callable=ttft_span.start_child,
                            op=f"{synthesizer_base_name}{CustomSentrySpans.SYNTHESIZER_CREATE_SPEECH}",
                        )
                maybe_synthesis_result: Optional[SynthesisResult] = None
                if isinstance(
                    self.conversation.synthesizer,
                    InputStreamingSynthesizer,
                ) and isinstance(agent_response_message.message, LLMToken):
                    logger.debug("Sending chunk to synthesizer")
                    await self.conversation.synthesizer.send_token_to_synthesizer(
                        message=agent_response_message.message,
                        chunk_size=self.chunk_size,
                    )
                else:
                    logger.debug("Synthesizing speech for message")
                    maybe_synthesis_result = await self.conversation.synthesizer.create_speech(
                        agent_response_message.message,
                        self.chunk_size,
                        is_first_text_chunk=self.is_first_text_chunk,
                        is_sole_text_chunk=agent_response_message.is_sole_text_chunk,
                    )
                if create_speech_span:
                    create_speech_span.finish()
                # For input streaming synthesizers, subsequent chunks are contained in the same SynthesisResult
                if isinstance(self.conversation.synthesizer, InputStreamingSynthesizer):
                    if not self.is_first_text_chunk:
                        maybe_synthesis_result = None
                    elif isinstance(agent_response_message.message, LLMToken):
                        maybe_synthesis_result = (
                            self.conversation.synthesizer.get_current_utterance_synthesis_result()
                        )
                if maybe_synthesis_result is not None:
                    synthesis_result = maybe_synthesis_result
                    synthesis_result.is_first = agent_response_message.is_first
                    if not synthesis_result.cached and synthesis_span:
                        synthesis_result.synthesis_total_span = synthesis_span
                        synthesis_result.ttft_span = ttft_span
                    self.produce_interruptible_agent_response_event_nonblocking(
                        (agent_response_message.message, synthesis_result),
                        is_interruptible=item.is_interruptible,
                        agent_response_tracker=item.agent_response_tracker,
                    )
                self.last_agent_response_tracker = item.agent_response_tracker
                if not isinstance(agent_response_message.message, SilenceMessage):
                    self.is_first_text_chunk = False
            except asyncio.CancelledError:
                pass

    class SynthesisResultsWorker(InterruptibleAgentResponseWorker):
        """Plays SynthesisResults from the output queue on the output device"""

        def __init__(
            self,
            input_queue: asyncio.Queue[
                InterruptibleAgentResponseEvent[
                    Tuple[Union[BaseMessage, EndOfTurn], Optional[SynthesisResult]]
                ]
            ],
            conversation: "StreamingConversation",
        ):
            super().__init__(input_queue=input_queue)
            self.input_queue = input_queue
            self.conversation = conversation
            self.last_transcript_message: Optional[Message] = None

        async def process(
            self,
            item: InterruptibleAgentResponseEvent[
                Tuple[Union[BaseMessage, EndOfTurn], Optional[SynthesisResult]]
            ],
        ):
            try:
                message, synthesis_result = item.payload
                if isinstance(message, EndOfTurn):
                    if self.last_transcript_message is not None:
                        self.last_transcript_message.is_end_of_turn = True
                    item.agent_response_tracker.set()
                    return
                assert synthesis_result is not None
                # create an empty transcript message and attach it to the transcript
                transcript_message = Message(
                    text="",
                    sender=Sender.BOT,
                    is_backchannel=isinstance(message, BotBackchannel),
                )
                if not isinstance(message, SilenceMessage):
                    self.conversation.transcript.add_message(
                        message=transcript_message,
                        conversation_id=self.conversation.id,
                        publish_to_events_manager=False,
                    )
                if isinstance(message, SilenceMessage):
                    logger.debug(f"Sending {message.trailing_silence_seconds} seconds of silence")
                elif isinstance(message, BotBackchannel):
                    logger.debug(f"Sending backchannel: {message}")
                message_sent, cut_off = await self.conversation.send_speech_to_output(
                    message.text,
                    synthesis_result,
                    item.interruption_event,
                    TEXT_TO_SPEECH_CHUNK_SIZE_SECONDS,
                    transcript_message=transcript_message,
                )
                # publish the transcript message now that it includes what was said during send_speech_to_output
                self.conversation.transcript.maybe_publish_transcript_event_from_message(
                    message=transcript_message,
                    conversation_id=self.conversation.id,
                )
                item.agent_response_tracker.set()
                logger.debug("Message sent: {}".format(message_sent))
                if cut_off:
                    self.conversation.agent.update_last_bot_message_on_cut_off(message_sent)
                self.last_transcript_message = transcript_message
            except asyncio.CancelledError:
                pass

    def __init__(
        self,
        output_device: OutputDeviceType,
        transcriber: BaseTranscriber[TranscriberConfig],
        agent: BaseAgent,
        synthesizer: BaseSynthesizer,
        speed_coefficient: float = 1.0,
        conversation_id: Optional[str] = None,
        per_chunk_allowance_seconds: float = PER_CHUNK_ALLOWANCE_SECONDS,
        events_manager: Optional[EventsManager] = None,
    ):
        self.id = conversation_id or create_conversation_id()
        self.output_device = output_device
        self.transcriber = transcriber
        self.agent = agent
        self.synthesizer = synthesizer
        self.synthesis_enabled = True

        self.interruptible_events: queue.Queue[InterruptibleEvent] = queue.Queue()
        self.interruptible_event_factory = self.QueueingInterruptibleEventFactory(conversation=self)
        self.agent.set_interruptible_event_factory(self.interruptible_event_factory)
        self.synthesis_results_queue: asyncio.Queue[
            InterruptibleAgentResponseEvent[
                Tuple[Union[BaseMessage, EndOfTurn], Optional[SynthesisResult]]
            ]
        ] = asyncio.Queue()
        self.filler_audio_queue: asyncio.Queue[InterruptibleAgentResponseEvent[FillerAudio]] = (
            asyncio.Queue()
        )
        self.state_manager = self.create_state_manager()
        self.transcriptions_worker = self.TranscriptionsWorker(
            input_queue=self.transcriber.output_queue,
            output_queue=self.agent.get_input_queue(),
            conversation=self,
            interruptible_event_factory=self.interruptible_event_factory,
        )
        self.agent.attach_conversation_state_manager(self.state_manager)
        self.agent_responses_worker = self.AgentResponsesWorker(
            input_queue=self.agent.get_output_queue(),
            output_queue=self.synthesis_results_queue,
            conversation=self,
            interruptible_event_factory=self.interruptible_event_factory,
        )
        self.actions_worker = None
        if self.agent.get_agent_config().actions:
            self.actions_worker = ActionsWorker(
                input_queue=self.agent.actions_queue,
                output_queue=self.agent.get_input_queue(),
                interruptible_event_factory=self.interruptible_event_factory,
                action_factory=self.agent.action_factory,
            )
            self.actions_worker.attach_conversation_state_manager(self.state_manager)
        self.synthesis_results_worker = self.SynthesisResultsWorker(
            input_queue=self.synthesis_results_queue,
            conversation=self,
        )
        self.filler_audio_worker = None
        self.filler_audio_config: Optional[FillerAudioConfig] = None
        if self.agent.get_agent_config().send_filler_audio:
            self.filler_audio_worker = self.FillerAudioWorker(
                input_queue=self.filler_audio_queue,
                conversation=self,
            )

        self.speed_coefficient = speed_coefficient
        self.speed_manager = SpeedManager(
            speed_coefficient=self.speed_coefficient,
        )
        self.transcriber.attach_speed_manager(self.speed_manager)
        self.agent.attach_speed_manager(self.speed_manager)

        self.events_manager = events_manager or EventsManager()
        self.events_task: Optional[asyncio.Task] = None
        self.per_chunk_allowance_seconds = per_chunk_allowance_seconds
        self.transcript = Transcript()
        self.transcript.attach_events_manager(self.events_manager)

        self.is_human_speaking = False
        self.active = False
        self.mark_last_action_timestamp()

        self.check_for_idle_task: Optional[asyncio.Task] = None
        self.check_for_idle_paused = False

        self.current_transcription_is_interrupt: bool = False

        self.initial_message_tracker = asyncio.Event()

        # tracing
        self.start_time: Optional[float] = None
        self.end_time: Optional[float] = None

        self.idle_time_threshold = (
            self.agent.get_agent_config().allowed_idle_time_seconds or ALLOWED_IDLE_TIME
        )

    def create_state_manager(self) -> ConversationStateManager:
        return ConversationStateManager(conversation=self)

    async def start(self, mark_ready: Optional[Callable[[], Awaitable[None]]] = None):
        self.transcriber.start()
        self.transcriptions_worker.start()
        self.agent_responses_worker.start()
        self.synthesis_results_worker.start()
        self.output_device.start()
        if self.filler_audio_worker is not None:
            self.filler_audio_worker.start()
        if self.actions_worker is not None:
            self.actions_worker.start()
        is_ready = await self.transcriber.ready()
        if not is_ready:
            raise Exception("Transcriber startup failed")
        if self.agent.get_agent_config().send_filler_audio:
            if not isinstance(self.agent.get_agent_config().send_filler_audio, FillerAudioConfig):
                self.filler_audio_config = FillerAudioConfig()
            else:
                self.filler_audio_config = typing.cast(
                    FillerAudioConfig,
                    self.agent.get_agent_config().send_filler_audio,
                )
            await self.synthesizer.set_filler_audios(self.filler_audio_config)

        self.agent.start()
        initial_message = self.agent.get_agent_config().initial_message
        if initial_message:
            asyncio_create_task_with_done_error_log(
                self.send_initial_message(initial_message),
            )
        else:
            self.initial_message_tracker.set()
        self.agent.attach_transcript(self.transcript)
        if mark_ready:
            await mark_ready()
        self.active = True
        self.check_for_idle_task = asyncio_create_task_with_done_error_log(
            self.check_for_idle(),
        )
        if len(self.events_manager.subscriptions) > 0:
            self.events_task = asyncio_create_task_with_done_error_log(
                self.events_manager.start(),
            )

    def set_check_for_idle_paused(self, paused: bool):
        logger.debug(f"Setting idle check paused to {paused}")
        if not paused:
            self.mark_last_action_timestamp()
        self.check_for_idle_paused = paused

    async def send_initial_message(
        self,
        initial_message: BaseMessage,
    ):
        # TODO: configure if initial message is interruptible
        delay = self.agent.get_agent_config().initial_message_delay
        if delay > 0:
            logger.info(f"Waiting {delay} seconds before initial message")
            await asyncio.sleep(delay)
        await self.send_single_message(
            message=initial_message,
            message_tracker=self.initial_message_tracker,
        )
        await self.initial_message_tracker.wait()

    async def action_on_idle(self):
        logger.debug("Conversation idle for too long, terminating")
        await self.mark_terminated(bot_disconnect=True)
        return

    async def check_for_idle(self):
        """Asks if human is still on the line if no activity is detected, and terminates the conversation if not."""
        await self.initial_message_tracker.wait()
        check_human_present_count = 0
        check_human_present_threshold = self.agent.get_agent_config().num_check_human_present_times
        while self.is_active():
            if (
                not self.check_for_idle_paused
            ) and time.time() - self.last_action_timestamp > self.idle_time_threshold:
                if check_human_present_count >= check_human_present_threshold:
                    # Stop the phone call after some retries to prevent infinitely long call where human is just silent.
                    await self.action_on_idle()
                await self.send_single_message(
                    message=BaseMessage(text=random.choice(CHECK_HUMAN_PRESENT_MESSAGE_CHOICES)),
                )
                check_human_present_count += 1
            # wait till the idle time would have passed the threshold if no action occurs
            await asyncio.sleep(ALLOWED_IDLE_TIME)

    async def send_single_message(
        self,
        message: BaseMessage,
        message_tracker: Optional[asyncio.Event] = None,
    ):
        agent_response_event = (
            self.interruptible_event_factory.create_interruptible_agent_response_event(
                AgentResponseMessage(message=message, is_sole_text_chunk=True),
                is_interruptible=False,
                agent_response_tracker=message_tracker,
            )
        )
        self.agent_responses_worker.consume_nonblocking(agent_response_event)
        self.agent_responses_worker.consume_nonblocking(
            self.interruptible_event_factory.create_interruptible_agent_response_event(
                AgentResponseMessage(message=EndOfTurn()),
                is_interruptible=True,
            ),
        )

    def receive_message(self, message: str):
        transcription = Transcription(
            message=message,
            confidence=1.0,
            is_final=True,
        )
        self.transcriptions_worker.consume_nonblocking(transcription)

    def receive_audio(self, chunk: bytes):
        self.transcriber.send_audio(chunk)

    def warmup_synthesizer(self):
        self.synthesizer.ready_synthesizer(self._get_synthesizer_chunk_size())

    def mark_last_action_timestamp(self):
        self.last_action_timestamp = time.time()

    def broadcast_interrupt(self):
        """Stops all inflight events and cancels all workers that are sending output

        Returns true if any events were interrupted - which is used as a flag for the agent (is_interrupt)
        """
        num_interrupts = 0
        while True:
            try:
                interruptible_event = self.interruptible_events.get_nowait()
                if not interruptible_event.is_interrupted():
                    if interruptible_event.interrupt():
                        logger.debug(
                            f"Interrupting event {type(interruptible_event.payload)} {interruptible_event.payload}",
                        )
                        num_interrupts += 1
            except queue.Empty:
                break
        self.agent.cancel_current_task()
        self.agent_responses_worker.cancel_current_task()
        if self.actions_worker:
            self.actions_worker.cancel_current_task()
        return num_interrupts > 0

    def is_interrupt(self, transcription: Transcription):
        return transcription.confidence >= (
            self.transcriber.get_transcriber_config().min_interrupt_confidence or 0
        )

    def _maybe_create_first_chunk_span(self, synthesis_result: SynthesisResult, message: str):
        first_chunk_span: Optional[Span] = None
        if synthesis_result.synthesis_total_span:
            synthesis_result.synthesis_total_span.set_tag("message_length", len(message))
        if synthesis_result.ttft_span:
            synthesis_result.ttft_span.set_tag("message_length", len(message))
            first_chunk_span = sentry_create_span(
                sentry_callable=synthesis_result.ttft_span.start_child,
                op=CustomSentrySpans.SYNTHESIS_GENERATE_FIRST_CHUNK,
            )
        return first_chunk_span

    def _track_first_chunk(self, first_chunk_span: Span, synthesis_result: SynthesisResult):
        complete_span_by_op(CustomSentrySpans.SYNTHESIS_TIME_TO_FIRST_TOKEN)
        first_chunk_span.finish()
        if synthesis_result.ttft_span:
            synthesis_result.ttft_span.finish()
        complete_span_by_op(CustomSentrySpans.LATENCY_OF_CONVERSATION)

    def _get_synthesizer_chunk_size(
        self,
        seconds_per_chunk: float = TEXT_TO_SPEECH_CHUNK_SIZE_SECONDS,
    ):
        return int(
            seconds_per_chunk
            * get_chunk_size_per_second(
                self.synthesizer.get_synthesizer_config().audio_encoding,
                self.synthesizer.get_synthesizer_config().sampling_rate,
            ),
        )

    async def send_speech_to_output(
        self,
        message: str,
        synthesis_result: SynthesisResult,
        stop_event: threading.Event,
        seconds_per_chunk: float,
        transcript_message: Optional[Message] = None,
        started_event: Optional[threading.Event] = None,
    ):
        """
        - Sends the speech chunk by chunk to the output device
          - update the transcript message as chunks come in (transcript_message is always provided for non filler audio utterances)
        - If the stop_event is set, the output is stopped
        - Sets started_event when the first chunk is sent

        Importantly, we rate limit the chunks sent to the output. For interrupts to work properly,
        the next chunk of audio can only be sent after the last chunk is played, so we send
        a chunk of x seconds only after x seconds have passed since the last chunk was sent.

        Returns the message that was sent up to, and a flag if the message was cut off
        """

        async def get_chunks(
            output_queue: asyncio.Queue[Optional[SynthesisResult.ChunkResult]],
            chunk_generator: AsyncGenerator[SynthesisResult.ChunkResult, None],
        ):
            try:
                async for chunk_result in chunk_generator:
                    await output_queue.put(chunk_result)
            except asyncio.CancelledError:
                pass
            finally:
                await output_queue.put(None)  # sentinel

        if self.transcriber.get_transcriber_config().mute_during_speech:
            logger.debug("Muting transcriber")
            self.transcriber.mute()
        message_sent = message
        cut_off = False
        chunk_size = self._get_synthesizer_chunk_size(seconds_per_chunk)
        chunk_idx = 0
        seconds_spoken = 0.0
        logger.debug(f"Start sending speech {message} to output")

        first_chunk_span = self._maybe_create_first_chunk_span(synthesis_result, message)
        chunk_queue: asyncio.Queue[Optional[SynthesisResult.ChunkResult]] = asyncio.Queue()
        get_chunks_task = asyncio_create_task_with_done_error_log(
            get_chunks(chunk_queue, synthesis_result.chunk_generator),
        )
        first = True
        while True:
            chunk_result = await chunk_queue.get()
            if chunk_result is None:
                break
            if first and first_chunk_span:
                self._track_first_chunk(first_chunk_span, synthesis_result)
            first = False
            start_time = time.time()
            speech_length_seconds = seconds_per_chunk * (len(chunk_result.chunk) / chunk_size)
            seconds_spoken = chunk_idx * seconds_per_chunk
            if stop_event.is_set():
                logger.debug(
                    "Interrupted, stopping text to speech after {} chunks".format(chunk_idx),
                )
                message_sent = synthesis_result.get_message_up_to(seconds_spoken)
                cut_off = True
                break
            if chunk_idx == 0:
                if started_event:
                    started_event.set()
            self.output_device.consume_nonblocking(chunk_result.chunk)
            end_time = time.time()
            await asyncio.sleep(
                max(
                    speech_length_seconds
                    - (end_time - start_time)
                    - self.per_chunk_allowance_seconds,
                    0,
                ),
            )
            self.mark_last_action_timestamp()
            chunk_idx += 1
            seconds_spoken += seconds_per_chunk
            if transcript_message:
                transcript_message.text = synthesis_result.get_message_up_to(seconds_spoken)
        get_chunks_task.cancel()
        if self.transcriber.get_transcriber_config().mute_during_speech:
            logger.debug("Unmuting transcriber")
            self.transcriber.unmute()
        if transcript_message:
            transcript_message.text = message_sent
            transcript_message.is_final = not cut_off
        if synthesis_result.synthesis_total_span:
            synthesis_result.synthesis_total_span.finish()
        return message_sent, cut_off

    async def mark_terminated(self, bot_disconnect: bool = False):
        self.active = False

    async def terminate(self):
        await self.mark_terminated()
        self.broadcast_interrupt()
        self.events_manager.publish_event(
            TranscriptCompleteEvent(
                conversation_id=self.id,
                transcript=self.transcript,
            ),
        )
        if self.check_for_idle_task:
            logger.debug("Terminating check_for_idle Task")
            self.check_for_idle_task.cancel()
        if self.events_manager and self.events_task:
            logger.debug("Terminating events Task")
            self.events_task.cancel()
            await self.events_manager.flush()
        logger.debug("Tearing down synthesizer")
        await self.synthesizer.tear_down()
        logger.debug("Terminating agent")
        if isinstance(self.agent, ChatGPTAgent) and self.agent.agent_config.vector_db_config:
            # Shutting down the vector db should be done in the agent's terminate method,
            # but it is done here because `vector_db.tear_down()` is async and
            # `agent.terminate()` is not async.
            logger.debug("Terminating vector db")
            await self.agent.vector_db.tear_down()
        self.agent.terminate()
        logger.debug("Terminating output device")
        self.output_device.terminate()
        logger.debug("Terminating speech transcriber")
        self.transcriber.terminate()
        logger.debug("Terminating transcriptions worker")
        self.transcriptions_worker.terminate()
        logger.debug("Terminating final transcriptions worker")
        self.agent_responses_worker.terminate()
        logger.debug("Terminating synthesis results worker")
        self.synthesis_results_worker.terminate()
        if self.filler_audio_worker is not None:
            logger.debug("Terminating filler audio worker")
            self.filler_audio_worker.terminate()
        if self.actions_worker is not None:
            logger.debug("Terminating actions worker")
            self.actions_worker.terminate()
        logger.debug("Successfully terminated")

    def is_active(self):
        return self.active



import time
from datetime import datetime
from typing import List, Literal, Optional

from pydantic.v1 import BaseModel, Field

from vocode.streaming.models.actions import ActionInput, ActionOutput
from vocode.streaming.models.events import ActionEvent, Event, EventType, Sender
from vocode.streaming.utils.events_manager import EventsManager


class EventLog(BaseModel):
    sender: Sender
    timestamp: float = Field(default_factory=time.time)

    def to_string(self, include_timestamp: bool = False) -> str:
        raise NotImplementedError

    def get_timestamp_string(self, start_timestamp: float) -> str:
        dt = datetime.fromtimestamp(self.timestamp - start_timestamp)
        return f"[{dt.strftime('%M:%S')}.{dt.microsecond // 10000:02}]"


class Message(EventLog):
    text: str
    is_final: bool = False
    is_backchannel: bool = False
    is_end_of_turn: bool = False

    def to_string(
        self,
        include_timestamp: bool = False,
        mark_human_backchannels_with_brackets: bool = False,
        include_sender: bool = True,
    ) -> str:
        text = self.text
        if not self.is_final and self.sender == Sender.BOT:
            text = f"{text}-"
        if self.is_backchannel and mark_human_backchannels_with_brackets:
            text = f"[{text}]"
        if include_timestamp:
            return f"{self.sender.name}: {text} ({self.timestamp})"
        elif include_sender:
            return f"{self.sender.name}: {text}"
        else:
            return text


class ActionStart(EventLog):
    sender: Sender = Sender.ACTION_WORKER
    action_type: str
    action_input: ActionInput

    def to_string(self, include_timestamp: bool = False, include_header: bool = True):
        main_string = self.action_input.action_config.action_attempt_to_string(self.action_input)
        if include_header:
            main_string = f"BOT_ACTION_START: {main_string}"
        if include_timestamp:
            return f"{main_string} ({self.timestamp})"
        return main_string


class ActionFinish(EventLog):
    sender: Sender = Sender.ACTION_WORKER
    action_type: str
    action_input: ActionInput
    action_output: ActionOutput

    def to_string(self, include_timestamp: bool = False, include_header: bool = True):
        main_string = self.action_input.action_config.action_result_to_string(
            self.action_input, self.action_output
        )
        if include_header:
            main_string = f"BOT_ACTION_FINISH: {main_string}"
        if include_timestamp:
            return f"{main_string} ({self.timestamp})"
        return main_string


ConferenceEventType = Literal[
    "participant_joined", "participant_left", "participant_unanswered", "voicemail"
]
ConferenceEventRole = Literal["primary", "transfer"]


class ConferenceEvent(EventLog):
    sender: Sender = Sender.CONFERENCE
    conference_event_type: ConferenceEventType
    conference_event_role: ConferenceEventRole
    participant_phone_number: str

    def to_string(self, include_timestamp: bool = False, include_sender: bool = True):
        if (
            self.conference_event_type == "participant_unanswered"
            or self.conference_event_type == "voicemail"
        ):
            msg_string = f"{self.conference_event_role.capitalize()} number ({self.participant_phone_number}) did not join the conference because they are busy"
        else:
            verb = "joined" if self.conference_event_type == "participant_joined" else "left"
            msg_string = f"{self.conference_event_role.capitalize()} number ({self.participant_phone_number}) {verb} the conference"
        if include_sender:
            msg_string = f"CONFERENCE_EVENT: {msg_string}"
        if include_timestamp:
            return f"{msg_string} ({self.timestamp})"
        return msg_string


class TranscriptEvent(Event, type=EventType.TRANSCRIPT):  # type: ignore
    text: str
    sender: Sender
    timestamp: float

    def to_string(self, include_timestamp: bool = False) -> str:
        if include_timestamp:
            return f"{self.sender.name}: {self.text} ({self.timestamp})"
        return f"{self.sender.name}: {self.text}"


class Transcript(BaseModel):
    event_logs: List[EventLog] = []
    start_time: float = Field(default_factory=time.time)
    events_manager: Optional[EventsManager] = None

    class Config:
        arbitrary_types_allowed = True

    def to_string(
        self, include_timestamps: bool = False, mark_human_backchannels_with_brackets: bool = False
    ) -> str:
        event_strings = []
        for event in self.event_logs:
            if isinstance(event, Message):
                event_string = event.to_string(
                    include_timestamp=False,
                    mark_human_backchannels_with_brackets=mark_human_backchannels_with_brackets,
                )
            else:
                event_string = event.to_string(include_timestamp=False)

            if include_timestamps:
                event_string = f"{event.get_timestamp_string(self.start_time)} {event_string}"
            event_strings.append(event_string)

        return "\n".join(event_strings)

    def attach_events_manager(self, events_manager: EventsManager):
        self.events_manager = events_manager

    def maybe_publish_transcript_event_from_message(self, message: Message, conversation_id: str):
        if self.events_manager is not None:
            self.events_manager.publish_event(
                TranscriptEvent(
                    text=message.text,
                    sender=message.sender,
                    timestamp=message.timestamp,
                    conversation_id=conversation_id,
                )
            )

    def add_message_from_props(
        self,
        text: str,
        sender: Sender,
        conversation_id: str,
        is_final: bool = False,
        is_backchannel: bool = False,
        publish_to_events_manager: bool = True,
    ):
        timestamp = time.time()
        message = Message(
            text=text,
            sender=sender,
            timestamp=timestamp,
            is_final=is_final,
            is_backchannel=is_backchannel,
        )
        self.event_logs.append(message)
        if publish_to_events_manager:
            self.maybe_publish_transcript_event_from_message(
                message=message, conversation_id=conversation_id
            )

    def add_message(
        self,
        message: Message,
        conversation_id: str,
        publish_to_events_manager: bool = True,
    ):
        self.event_logs.append(message)
        if publish_to_events_manager:
            self.maybe_publish_transcript_event_from_message(
                message=message, conversation_id=conversation_id
            )

    def add_human_message(self, text: str, conversation_id: str, is_backchannel: bool = False):
        self.add_message_from_props(
            text=text,
            sender=Sender.HUMAN,
            conversation_id=conversation_id,
            is_backchannel=is_backchannel,
        )

    def add_bot_message(self, text: str, conversation_id: str, is_final: bool = False):
        self.add_message_from_props(
            text=text,
            sender=Sender.BOT,
            conversation_id=conversation_id,
            is_final=is_final,
        )

    def get_last_user_message(self):
        for idx, message in enumerate(self.event_logs[::-1]):
            if message.sender == Sender.HUMAN:
                return -1 * (idx + 1), message.to_string()

    def add_action_start_log(self, action_input: ActionInput, conversation_id: str):
        timestamp = time.time()
        self.event_logs.append(
            ActionStart(
                action_input=action_input,
                action_type=action_input.action_config.type,
                timestamp=timestamp,
            )
        )
        if self.events_manager is not None:
            self.events_manager.publish_event(
                ActionEvent(
                    action_input=action_input.dict(),
                    conversation_id=conversation_id,
                )
            )

    def add_action_finish_log(
        self,
        action_input: ActionInput,
        action_output: ActionOutput,
        conversation_id: str,
    ):
        timestamp = time.time()
        self.event_logs.append(
            ActionFinish(
                action_input=action_input,
                action_output=action_output,
                action_type=action_output.action_type,
                timestamp=timestamp,
            )
        )
        if self.events_manager is not None:
            self.events_manager.publish_event(
                ActionEvent(
                    action_input=action_input.dict(),
                    action_output=action_output.dict(),
                    conversation_id=conversation_id,
                )
            )

    def update_last_bot_message_on_cut_off(self, text: str):
        # TODO: figure out what to do for the event
        for event_log in reversed(self.event_logs):
            if isinstance(event_log, Message) and event_log.sender == Sender.BOT:
                event_log.text = text
                break

    def was_last_message_interrupted(self):
        bot_messages = [
            message
            for message in self.event_logs
            if isinstance(message, Message) and message.sender == Sender.BOT
        ]
        if len(bot_messages) > 0:
            last_bot_message = bot_messages[-1]
            return not last_bot_message.is_final or not last_bot_message.is_end_of_turn
        return False


class TranscriptCompleteEvent(Event, type=EventType.TRANSCRIPT_COMPLETE):  # type: ignore
    transcript: Transcript



from typing import Any, List, Tuple

from pydantic.v1 import BaseModel as Pydantic1BaseModel


class BaseModel(Pydantic1BaseModel):
    def __init__(self, **data):
        for key, value in data.items():
            if isinstance(value, dict):
                if (
                    "type" in value and key != "action_trigger"
                ):  # TODO: this is a quick workaround until we get a vocode object version of action trigger (ajay has approved it)
                    data[key] = TypedModel.parse_obj(value)
            if isinstance(value, list):
                for i, v in enumerate(value):
                    if isinstance(v, dict):
                        if "type" in v:
                            value[i] = TypedModel.parse_obj(v)
        super().__init__(**data)


# Adapted from https://github.com/pydantic/pydantic/discussions/3091
class TypedModel(BaseModel):
    _subtypes_: List[Tuple[Any, Any]] = []

    def __init_subclass__(cls, type=None):  # type: ignore
        cls._subtypes_.append((type, cls))

    @classmethod
    def get_cls(_cls, type):
        for t, cls in _cls._subtypes_:
            if t == type:
                return cls
        raise ValueError(f"Unknown type {type}")

    @classmethod
    def get_type(_cls, cls_name):
        for t, cls in _cls._subtypes_:
            if cls.__name__ == cls_name:
                return t
        raise ValueError(f"Unknown class {cls_name}")

    @classmethod
    def parse_obj(cls, obj):
        data_type = obj.get("type")
        if data_type is None:
            raise ValueError(f"type is required for {cls.__name__}")

        sub = cls.get_cls(data_type)
        if sub is None:
            raise ValueError(f"Unknown type {data_type}")
        return sub(**obj)

    def _iter(self, **kwargs):
        yield "type", self.get_type(self.__class__.__name__)
        yield from super()._iter(**kwargs)

    @property
    def type(self):
        return self.get_type(self.__class__.__name__)



from typing import AsyncGenerator

from loguru import logger
from pydantic.v1 import BaseModel, parse_obj_as
from redis.asyncio import Redis

from vocode.streaming.utils.redis import initialize_redis
from vocode.streaming.utils.singleton import Singleton


class RedisMessage(BaseModel):
    type: str


class RedisConversationMessageQueue(Singleton):
    def __init__(self):
        self.redis: Redis = initialize_redis()

    async def publish(self, conversation_id: str, message: RedisMessage):
        logger.info(f"Publishing message to {conversation_id}")
        await self.redis.xadd(f"{conversation_id}:stream", message.dict())  # type: ignore

    async def wait_for_messages(
        self, conversation_id: str, timeout_seconds: int = 20
    ) -> AsyncGenerator[RedisMessage, None]:
        logger.info(f"Waiting for message from {conversation_id}")
        streams = await self.redis.xread(
            {f"{conversation_id}:stream": "$"}, block=timeout_seconds * 1000
        )
        for _, stream in streams:  # stream_id, stream
            for _, message in stream:  # timestamp, message
                yield parse_obj_as(RedisMessage, message)  # type: ignore

    async def clear_stream(self, conversation_id: str):
        await self.redis.delete(f"{conversation_id}:stream")



import re
from typing import List, Optional

from vocode.streaming.models.actions import ActionConfig, PhraseBasedActionTrigger


def matches_phrase_trigger(
    message: str,
    action_configs: List[ActionConfig],
) -> Optional[str]:
    cleaned = re.sub(r"[^\w\s]", "", message.lower())
    for action_config in action_configs:
        if not isinstance(action_config.action_trigger, PhraseBasedActionTrigger):
            continue

        for phrase_trigger in action_config.action_trigger.config.phrase_triggers:
            lowered = phrase_trigger.phrase.lower()
            for condition in phrase_trigger.conditions:
                if condition == "phrase_condition_type_contains" and lowered in cleaned:
                    return action_config.type
    return None



from enum import Enum

import speech_recognition as sr
from pydub import AudioSegment

from vocode import getenv
from vocode.turn_based.transcriber.base_transcriber import BaseTranscriber


class SpeechRecognitionAPI(Enum):
    SPHINX = "sphinx"
    GOOGLE = "google"
    GOOGLE_CLOUD = "google_cloud"
    WIT = "wit"
    BING = "bing"
    AZURE = "azure"
    HOUNDIFY = "houndify"
    IBM = "ibm"


class SpeechRecognitionTranscriber(BaseTranscriber):
    def __init__(self, api: SpeechRecognitionAPI = SpeechRecognitionAPI.GOOGLE):
        self.api = api

    def transcribe(self, audio_segment: AudioSegment) -> str:
        audio_data = audio_segment.raw_data
        audio = sr.AudioData(
            audio_data,
            sample_rate=audio_segment.frame_rate,
            sample_width=audio_segment.sample_width,
        )

        r = sr.Recognizer()

        try:
            if self.api == SpeechRecognitionAPI.GOOGLE:
                api_key = getenv("GOOGLE_SPEECH_RECOGNITION_API_KEY")
                if not api_key:
                    text = r.recognize_google(audio)
                else:
                    text = r.recognize_google(audio, key=api_key)

            elif self.api == SpeechRecognitionAPI.SPHINX:
                # do note that sphinx requires PocketSphinx to be installed
                text = r.recognize_sphinx(audio)

            elif self.api == SpeechRecognitionAPI.GOOGLE_CLOUD:
                credentials_json = getenv("GOOGLE_CLOUD_SPEECH_CREDENTIALS")
                if not credentials_json:
                    raise ValueError("Google Cloud Speech credentials not provided")
                text = r.recognize_google_cloud(audio, credentials_json=credentials_json)

            elif self.api == SpeechRecognitionAPI.WIT:
                api_key = getenv("WIT_AI_API_KEY")
                if not api_key:
                    raise ValueError("Wit.ai API key not provided")
                text = r.recognize_wit(audio, key=api_key)

            elif self.api == SpeechRecognitionAPI.BING:
                api_key = getenv("BING_API_KEY")
                if not api_key:
                    raise ValueError("Bing API key not provided")
                text = r.recognize_bing(audio, key=api_key)

            elif self.api == SpeechRecognitionAPI.AZURE:
                api_key = getenv("AZURE_SPEECH_KEY")
                region = getenv("AZURE_SPEECH_REGION")
                if not api_key:
                    raise ValueError("Azure Speech API key not provided")
                if not region:
                    raise ValueError("Azure Speech region not provided")
                text = r.recognize_azure(audio, key=api_key, location=region)[0]

            elif self.api == SpeechRecognitionAPI.HOUNDIFY:
                client_id = getenv("HOUNDIFY_CLIENT_ID")
                client_key = getenv("HOUNDIFY_CLIENT_KEY")
                if not client_id or not client_key:
                    raise ValueError("Houndify client ID or key not provided")
                text = r.recognize_houndify(audio, client_id=client_id, client_key=client_key)

            elif self.api == SpeechRecognitionAPI.IBM:
                username = getenv("IBM_USERNAME")
                password = getenv("IBM_PASSWORD")
                if not username or not password:
                    raise ValueError("IBM Speech to Text username or password not provided")
                text = r.recognize_ibm(audio, username=username, password=password)

            else:
                raise ValueError(f"Unsupported API: {self.api}")
        except sr.UnknownValueError:
            raise sr.UnknownValueError("Speech Recognition could not understand audio")
        except sr.RequestError as e:
            raise sr.RequestError(f"Could not request results from Speech Recognition service.")

        return text



from typing import Optional
from xml.etree import ElementTree

from pydub import AudioSegment
from regex import D

from vocode import getenv
from vocode.turn_based.synthesizer.base_synthesizer import BaseSynthesizer

DEFAULT_SAMPLING_RATE = 22050
DEFAULT_VOICE_NAME = "en-US-AriaNeural"
DEFAULT_PITCH = 0
DEFAULT_RATE = 15

NAMESPACES = {
    "mstts": "https://www.w3.org/2001/mstts",
    "": "https://www.w3.org/2001/10/synthesis",
}

ElementTree.register_namespace("", NAMESPACES[""])
ElementTree.register_namespace("mstts", NAMESPACES["mstts"])


class AzureSynthesizer(BaseSynthesizer):
    def __init__(
        self,
        sampling_rate: int = DEFAULT_SAMPLING_RATE,
        voice_name: str = DEFAULT_VOICE_NAME,
        pitch: int = DEFAULT_PITCH,
        rate: int = DEFAULT_RATE,
        api_key: Optional[str] = None,
        region: Optional[str] = None,
    ):
        import azure.cognitiveservices.speech as speechsdk

        self.speechsdk = speechsdk

        self.sampling_rate = sampling_rate
        speech_config = self.speechsdk.SpeechConfig(
            subscription=getenv("AZURE_SPEECH_KEY", api_key),
            region=getenv("AZURE_SPEECH_REGION", region),
        )
        if self.sampling_rate == 44100:
            speech_config.set_speech_synthesis_output_format(
                self.speechsdk.SpeechSynthesisOutputFormat.Raw44100Hz16BitMonoPcm
            )
        if self.sampling_rate == 48000:
            speech_config.set_speech_synthesis_output_format(
                self.speechsdk.SpeechSynthesisOutputFormat.Raw48Khz16BitMonoPcm
            )
        if self.sampling_rate == 24000:
            speech_config.set_speech_synthesis_output_format(
                self.speechsdk.SpeechSynthesisOutputFormat.Raw24Khz16BitMonoPcm
            )
        if self.sampling_rate == 22050:
            speech_config.set_speech_synthesis_output_format(
                self.speechsdk.SpeechSynthesisOutputFormat.Raw22050Hz16BitMonoPcm
            )
        elif self.sampling_rate == 16000:
            speech_config.set_speech_synthesis_output_format(
                self.speechsdk.SpeechSynthesisOutputFormat.Raw16Khz16BitMonoPcm
            )
        elif self.sampling_rate == 8000:
            speech_config.set_speech_synthesis_output_format(
                self.speechsdk.SpeechSynthesisOutputFormat.Raw8Khz16BitMonoPcm
            )

        self.synthesizer = self.speechsdk.SpeechSynthesizer(
            speech_config=speech_config, audio_config=None
        )
        self.voice_name = voice_name
        self.pitch = pitch
        self.rate = rate

    def create_ssml(self, message: str) -> str:
        ssml_root = ElementTree.fromstring(
            '<speak version="1.0" xmlns="https://www.w3.org/2001/10/synthesis" xml:lang="en-US"></speak>'
        )
        voice = ElementTree.SubElement(ssml_root, "voice")
        voice.set("name", self.voice_name)
        voice_root = voice
        prosody = ElementTree.SubElement(voice_root, "prosody")
        prosody.set("pitch", f"{self.pitch}%")
        prosody.set("rate", f"{self.rate}%")
        prosody.text = message.strip()
        return ElementTree.tostring(ssml_root, encoding="unicode")

    def synthesize(self, text) -> AudioSegment:
        result = self.synthesizer.speak_ssml(self.create_ssml(text))
        if result.reason == self.speechsdk.ResultReason.SynthesizingAudioCompleted:
            return AudioSegment(
                result.audio_data,
                sample_width=2,
                frame_rate=self.sampling_rate,
                channels=1,
            )
        else:
            raise Exception("Could not synthesize audio")



from loguru import logger


def ensure_punkt_installed():
    try:
        from nltk.data import find

        find("tokenizers/punkt")
    except LookupError:
        from nltk import download

        # If not installed, download 'punkt'
        logger.info("Downloading 'punkt' tokenizer...")
        download("punkt")
        logger.info("'punkt' tokenizer downloaded successfully.")



import asyncio
from enum import Enum
from typing import Annotated, Generic, List, Literal, Optional, TypeVar, Union

from pydantic.v1 import BaseModel, Field

from vocode.streaming.models.message import BaseMessage
from vocode.streaming.models.model import TypedModel

TriggerType = Literal["action_trigger_function_call", "action_trigger_phrase_based"]


class ActionTriggerConfig(BaseModel):
    pass


class _ActionTrigger(BaseModel):
    type: TriggerType
    config: ActionTriggerConfig


class FunctionCallActionTriggerConfig(ActionTriggerConfig):
    pass


class FunctionCallActionTrigger(_ActionTrigger):
    type: Literal["action_trigger_function_call"] = "action_trigger_function_call"
    config: FunctionCallActionTriggerConfig = Field(default_factory=FunctionCallActionTriggerConfig)


PhraseConditionType = Literal["phrase_condition_type_contains"]


class PhraseTrigger(BaseModel):
    phrase: str
    conditions: List[PhraseConditionType]


class PhraseBasedActionTriggerConfig(ActionTriggerConfig):
    phrase_triggers: List[PhraseTrigger]


class PhraseBasedActionTrigger(_ActionTrigger):
    type: Literal["action_trigger_phrase_based"] = "action_trigger_phrase_based"
    config: PhraseBasedActionTriggerConfig


ActionTrigger = Annotated[
    Union[FunctionCallActionTrigger, PhraseBasedActionTrigger],
    Field(discriminator="type"),
]


class ActionType(str, Enum):
    BASE = "action_base"
    NYLAS_SEND_EMAIL = "action_nylas_send_email"
    WAIT = "action_wait"
    RECORD_EMAIL = "action_record_email"
    END_CONVERSATION = "action_end_conversation"
    EXECUTE_EXTERNAL_ACTION = "action_external"

    TRANSFER_CALL = "action_transfer_call"
    DTMF = "action_dtmf"


ParametersType = TypeVar("ParametersType", bound=BaseModel)

ACTION_STARTED_FORMAT_STRING = "!STARTING ACTION {action_name} WITH PARAMETERS {action_params}!"
ACTION_FINISHED_FORMAT_STRING = "!ACTION {action_name} FINISHED WITH OUTPUT {action_output}!"


class ActionConfig(TypedModel, type=ActionType.BASE):  # type: ignore
    action_trigger: ActionTrigger = FunctionCallActionTrigger(type="action_trigger_function_call")

    def action_attempt_to_string(self, input: "ActionInput") -> str:
        return ACTION_STARTED_FORMAT_STRING.format(
            action_name=self.type,
            action_params=input.params.json(),
        )

    def action_result_to_string(self, input: "ActionInput", output: "ActionOutput") -> str:
        return ACTION_FINISHED_FORMAT_STRING.format(
            action_name=self.type,
            action_output=output.response.json(),
        )


class ActionInput(BaseModel, Generic[ParametersType]):
    action_config: ActionConfig
    conversation_id: str
    params: ParametersType
    user_message_tracker: Optional[asyncio.Event] = None

    class Config:
        arbitrary_types_allowed = True


class FunctionFragment(BaseModel):
    name: str
    arguments: str


class FunctionCall(BaseModel):
    name: str
    arguments: str


class EndOfTurn(BaseModel):
    pass


class VonagePhoneConversationActionInput(ActionInput[ParametersType]):
    vonage_uuid: str


class TwilioPhoneConversationActionInput(ActionInput[ParametersType]):
    twilio_sid: str


ResponseType = TypeVar("ResponseType", bound=BaseModel)


class ActionOutput(BaseModel, Generic[ResponseType]):
    action_type: str
    canned_response: Optional[BaseMessage] = None
    response: ResponseType


ExternalActionProcessingMode = Literal["muted"]
